#
# Parses a CellML 1.0 or 1.1 document.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import collections
import warnings

from lxml import etree

import myokit
import myokit.formats.mathml
import myokit.formats.cellml as cellml
import myokit.formats.cellml.v1

from myokit.formats.xml import split


def parse_file(path):
    """
    Parses a CellML 1.0 or 1.1 model at the given path and returns a
    :class:`myokit.formats.cellml.v1.Model`.

    Raises a :class:`CellMLParsingError` if anything goes wrong.

    For notes about CellML 1.0/1.1 support, see
    :class:`myokit.formats.cellml.v1.Model`.
    """
    return CellMLParser().parse_file(path)


def parse_string(text):
    """
    Parses a CellML 1.0 or 1.1 model from the given string and returns a
    :class:`myokit.formats.cellml.v1.Model`.

    Raises a :class:`CellMLParsingError` if anything goes wrong.

    For notes about CellML 1.0/1.1 support, see
    :class:`myokit.formats.cellml.v1.Model`.
    """
    return CellMLParser().parse_string(text)


class CellMLParsingError(myokit.ImportError):
    """
    Raised if an error occurs during CellML parsing.

    The argument ``element`` can be used to pass in an element that caused the
    error.
    """
    def __init__(self, message, element=None):
        if element is not None:
            try:    # pragma: no cover
                line = str(element.sourceline)
                message = 'Error on line ' + line + '. ' + message
            except AttributeError:
                pass
        super().__init__(message)


class CellMLParser:
    """
    Parses CellML 1.0 and 1.1 documents, and performs (partial) validation of a
    CellML document.

    For notes about CellML 1.0/1.1 support, see
    :class:`myokit.formats.cellml.v1.Model`.
    """

    # Note on assignment form, used in errors
    _dae_message = (
        'Each equation must have either a single variable or a derivative of a'
        ' single variable as its left-hand side, and each variable may only'
        ' appear on a left-hand side once.')

    def _check_allowed_content(
            self, element, children, attributes, name=None, math=False):
        """
        Scans ``element`` and raises an exception if any unlisted CellML
        children are found, if any unlisted null-namespace attributes are
        found, or if the element contains text.

        With ``math=False`` (default), the method also checks that no MathML
        elements are present. With ``math=True`` only a MathML ``<math>``
        element is allowed.
        """
        # Check for text inside this tag
        # For mixed-context xml this checks only up until the first child tag.
        if element.text is not None:
            if element.text.strip():
                raise CellMLParsingError(
                    'Text found in ' + self._tag(element, name) + '.', element)

        # Not extension namespaces
        not_ext_ns = (
            self._ns, cellml.NS_CMETA, cellml.NS_RDF, cellml.NS_MATHML)

        # Check child elements
        allowed = set([self._join(x) for x in children])
        allowed.add(self._join('RDF', cellml.NS_RDF))
        if math:
            allowed.add(self._join('math', cellml.NS_MATHML))

        # Check if only contains allowed child elements
        for child in element:
            # Check for trailing text
            if child.tail is not None and child.tail.strip():
                raise CellMLParsingError(
                    'Text found in ' + self._tag(element, name)
                    + ' (after ' + self._tag(child) + ' element).',
                    child)

            # Check if allowed
            if str(child.tag) in allowed:
                continue

            # Check if in an extension element
            ns = split(child.tag)[0]
            if ns in not_ext_ns:
                raise CellMLParsingError(
                    'Unexpected content type in ' + self._tag(element, name)
                    + ', found element of type ' + self._tag(child) + '.',
                    child)
            else:
                # Check if CellML appearing in non-CellML elements
                # But leave checking inside MathML for later
                self._check_for_cellml_in_extensions(child)

        # Check attributes
        allowed = set(attributes)
        cmeta_id = self._join('id', cellml.NS_CMETA)
        for key in element.attrib:
            # Cmeta id is always allowed
            if key == cmeta_id:
                continue

            # Extension namespaces are always allowed
            ns, at = split(key)
            if ns is not None and ns not in not_ext_ns:
                continue

            # No namespace, then must be in allowed list
            if key not in allowed:
                key = self._item(ns, at)
                raise CellMLParsingError(
                    'Unexpected attribute ' + key + ' found in '
                    + self._tag(element, name) + '.', element)

    def _check_for_cellml_in_extensions(self, element):
        """
        Checks that a non-CellML element does not contain CellML elements or
        attributes.
        """
        # Check if this element contains CellML attributes
        for key in element.attrib:
            ns, at = split(key)
            if ns == self._ns:
                raise CellMLParsingError(
                    'CellML attribute ' + self._item(ns, at) + ' found'
                    ' in extension element ' + self._tag(element)
                    + ' (2.4.3).', element)

        # Check if this element has CellML children
        for child in element:
            if split(child.tag)[0] == self._ns:
                raise CellMLParsingError(
                    'CellML element ' + self._tag(child) + ' found inside'
                    ' extension element ' + self._tag(element) + ' (2.4.3).',
                    child)

            # Recurse into children
            self._check_for_cellml_in_extensions(child)

    def _check_cmeta_id(self, element):
        """
        Checks and returns the ``cmeta_id`` of a given ``element``.

        If present, the well-formedness and uniqueness of the cmeta id is
        tested.
        """
        # Find attribute
        cmeta_id = element.attrib.get(self._join('id', cellml.NS_CMETA), None)
        if cmeta_id is None:
            return None

        # Check well-formedness
        if not cmeta_id:
            raise CellMLParsingError(
                'If present, a cmeta:id must be a non-empty string.', element)

        # Check uniqueness
        if cmeta_id in self._cmeta_ids:
            raise CellMLParsingError(
                'Duplicate cmeta:id "' + cmeta_id + '" (8.5.1).', element)

        # Store and return
        self._cmeta_ids.add(cmeta_id)
        return cmeta_id

    def flatten(self, element):
        """
        Converts an element tree to a plain text string.
        """
        import re
        import textwrap

        # Gather all text
        def gather(element, text=[]):
            """ Flatten and gather text. """
            if element.text is not None:
                text.append(element.text)
            for child in element:
                gather(child, text)
            if element.tail is not None:
                text.append(element.tail)
            return text

        # Gather text, replace whitespace
        text = ' '.join(gather(element)).strip()

        # Replace all whitespace except newlines with a single space
        text = re.sub(r'[ \t\f\r]+', ' ', text)

        # Remove leading/trailing spaces
        text = re.sub(r'[ \t\f\r]*[\n]+[ \t\f\r]*', '\n', text)

        # Remove double line-breaks
        text = re.sub(r'[\n][\n]+', '\n\n', text)

        # Replace non-ascii characters
        text = text.encode('ascii', errors='ignore').decode()

        # Space text so that it's readable(ish)
        lines = []
        for line in text.splitlines():
            if line:
                lines.extend(textwrap.wrap(line, width=75))
            else:
                lines.append('')
        text = '\n'.join(lines)

        return text

    def _item(self, ns, item):
        """
        Given a namespace ``ns`` and an element or attribute ``item``, this
        method will return a nicely formatted combination, e.g. ``item`` for
        the null namespace, ``cellml:item`` for a namespace with a known
        prefix, or ``{ns}item`` for a namespace without a known prefix.

        See also :meth:`_tag`.
        """
        if ns is None:
            return item
        elif ns == self._ns:
            return 'cellml:' + item
        elif ns == cellml.NS_MATHML:
            return 'mathml:' + item
        elif ns == cellml.NS_RDF:
            return 'rdf:' + item
        elif ns == cellml.NS_CMETA:
            return 'cmeta:' + item
        else:
            return '{' + ns + '}' + item

    def _join(self, element, namespace=None):
        """
        Joins a ``namespace`` and an ``element`` string into a single string.
        """
        if namespace is None:
            namespace = self._ns
        return '{' + namespace + '}' + element

    def parse(self, root):
        """
        Parses and validates a CellML document rooted in the given elementtree
        element.
        """
        # Parse model, using temporary state properties
        self._ns = None
        self._cmeta_ids = set()
        self._free_vars = set()
        try:
            return self._parse_model(root)
        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e))
        except myokit.formats.mathml.MathMLError as e:
            raise CellMLParsingError(str(e))
        finally:
            del self._ns, self._cmeta_ids, self._free_vars

    def parse_file(self, path):
        """
        Parses and validates the CellML file at ``path`` and returns a CellML
        Model.
        """
        # Read file
        try:
            parser = etree.XMLParser(remove_comments=True)
            tree = etree.parse(path, parser=parser)
        except Exception as e:
            raise CellMLParsingError('Unable to parse XML: ' + str(e))

        # Parse content
        return self.parse(tree.getroot())

    def parse_string(self, text):
        """
        Parses and validates the CellML XML in the string ``text`` and returns
        a CellML model.
        """
        # Read string
        import xml.etree.ElementTree as etree   # lxml makes this difficult
        try:
            root = etree.fromstring(text)
        except Exception as e:
            raise CellMLParsingError('Unable to parse XML: ' + str(e))

        # Parse content
        return self.parse(root)

    def _parse_component(self, element, model):
        """
        Parses a component ``element`` and adds it to the given ``model``.
        """
        # Check name is present
        try:
            name = element.attrib['name']
        except KeyError:
            raise CellMLParsingError(
                'Component element must have a name attribute (3.4.2.1).',
                element)

        # Create component (validates name and checks uniqueness)
        try:
            component = model.add_component(name)

            # Check cmeta id
            cmeta_id = self._check_cmeta_id(element)
            if cmeta_id:
                component.set_cmeta_id(cmeta_id)

        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e), element)

        # Check for reactions
        reaction = element.find(self._join('reaction'))
        if reaction is not None:
            raise CellMLParsingError('Reactions are not supported.', reaction)

        # Check allowed content
        self._check_allowed_content(
            element, ['units', 'variable'], ['name'], name, math=True)

        # Create component units
        for child in self._sort_units(element, model):
            self._parse_units(child, component)

        # Create variables and set interfaces
        for child in element.findall(self._join('variable')):
            self._parse_variable(child, component)

    def _parse_connection(self, element, model, connected):
        """
        Parses a connection ``element``.

        A set of tuples representing connected components is passed in as
        ``connected``. Each pair of components should be represented twice in
        this set; once in each order.
        """
        # Check exactly one map_components is present
        map_components = element.findall(self._join('map_components'))
        if len(map_components) != 1:
            raise CellMLParsingError(
                'A connection must contain exactly one map_components element,'
                ' found ' + str(len(map_components)) + ' (3.4.4.1).', element)

        # Check at least one map_variables is present
        map_variables = element.findall(self._join('map_variables'))
        if len(map_variables) < 1:
            raise CellMLParsingError(
                'A connection must contain at least one map_variables element'
                ' (3.4.4.1).', element)

        # Check cmeta id
        self._check_cmeta_id(element)

        # Check allowed content
        self._check_allowed_content(
            element, ['map_components', 'map_variables'], [])

        # Parse map_components element
        c1, c2 = self._parse_connection_map_components(
            map_components[0], model, connected)

        # Parse map_variables elements
        for child in map_variables:
            self._parse_connection_map_variables(child, c1, c2)

    def _parse_connection_map_components(self, element, model, connected):
        """
        Parses a map_components ``element``.
        """
        # Get component_1 and component_2
        try:
            c1 = element.attrib['component_1']
        except KeyError:
            raise CellMLParsingError(
                'A map_components element must define a component_1 attribute'
                ' (3.4.5.1).', element)
        try:
            c2 = element.attrib['component_2']
        except KeyError:
            raise CellMLParsingError(
                'A map_components element must define a component_2 attribute'
                ' (3.4.5.1).', element)

        # Check components are different
        if c1 == c2:
            raise CellMLParsingError(
                'The component_1 and component_2 attributes in a'
                ' map_components element must be different, got "' + str(c1)
                + '" twice (3.4.5.4).', element)

        # Get components
        try:
            c1 = model.component(c1)
        except KeyError:
            raise CellMLParsingError(
                'A map_components component_1 attribute must refer to a'
                ' component in the current model, got "' + str(c1)
                + '" (3.4.5.2).', element)
        try:
            c2 = model.component(c2)
        except KeyError:
            raise CellMLParsingError(
                'A map_components component_2 attribute must refer to a'
                ' component in the current model, got "' + str(c2)
                + '" (3.4.5.3).', element)

        # Check components are not yet connected
        if (c1, c2) in connected:
            raise CellMLParsingError(
                'Each connection in a model must connect a unique pair of'
                ' components, found multiple for "' + c1.name() + '" and "'
                + c2.name() + '" (3.4.5.4).', element)
        connected.add((c1, c2))
        connected.add((c2, c1))

        # Check cmeta id
        self._check_cmeta_id(element)

        # Check allowed content
        self._check_allowed_content(
            element, [], ['component_1', 'component_2'])

        # Return components
        return c1, c2

    def _parse_connection_map_variables(self, element, c1, c2):
        """
        Parses a map_variables ``element`` for a connection with components
        ``c1`` and ``c2``.
        """
        # Get variable_1 and variable_2
        try:
            v1 = element.attrib['variable_1']
        except KeyError:
            raise CellMLParsingError(
                'A map_variables element must define a variable_1 attribute'
                ' (3.4.6.1).', element)
        try:
            v2 = element.attrib['variable_2']
        except KeyError:
            raise CellMLParsingError(
                'A map_variables element must define a variable_2 attribute'
                ' (3.4.6.1).', element)

        # Get variables
        try:
            v1 = c1.variable(v1)
        except KeyError:
            raise CellMLParsingError(
                'A map_variables variable_1 attribute must refer to a'
                ' variable in component_1, got "' + str(v1) + '" (3.4.6.2).',
                element)
        try:
            v2 = c2.variable(v2)
        except KeyError:
            raise CellMLParsingError(
                'A map_variables variable_2 attribute must refer to a'
                ' variable in component_2, got "' + str(v1) + '" (3.4.6.3).',
                element)

        # Connect variables
        model = c1.model()
        try:
            model.add_connection(v1, v2)
        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e), element)

        # Check cmeta id
        self._check_cmeta_id(element)

        # Check allowed content
        self._check_allowed_content(element, [], ['variable_1', 'variable_2'])

    def _parse_documentation(self, element, model):
        """
        Parses an old-fashion <documentation> element (by flattening the
        content and adding it to the model's ``documentation`` meta data.
        """
        text = self.flatten(element)
        if text:
            if 'documentation' in model.meta:
                model.meta['documentation'] += '\n\n' + text
            else:
                model.meta['documentation'] = text

    def _parse_group(self, element, model):
        """
        Parses a group ``element`` and updates components in the given
        ``model``.
        """

        # Check cmeta id
        self._check_cmeta_id(element)

        # Check contains at least one component_ref and relationship_ref
        if element.find(self._join('component_ref')) is None:
            raise CellMLParsingError(
                'Group must contain at least one component_ref element'
                ' (6.4.1.1).', element)
        if element.find(self._join('relationship_ref')) is None:
            raise CellMLParsingError(
                'Group must contain at least one relationship_ref element'
                ' (6.4.1.1).', element)

        # Check allowed content
        self._check_allowed_content(
            element, ['relationship_ref', 'component_ref'], [])

        # Parse relationship refs
        relationships = set()
        for child in element.findall(self._join('relationship_ref')):
            self._parse_group_relationship_ref(child, relationships)

        # Parse component refs
        for child in element.findall(self._join('component_ref')):
            self._parse_group_component_ref(child, model, relationships)

    def _parse_group_component_ref(
            self, element, model, relationships, parent=None):
        """
        Parses a component_ref ``element``, using the information about its
        groups ``relationships``.

        If this component_ref is nested inside another component_ref, set the
        parent component as ``parent``.
        """
        # Check cmeta id
        self._check_cmeta_id(element)

        # Get component attribute
        try:
            component = element.attrib['component']
        except KeyError:
            raise CellMLParsingError(
                'A component_ref must define a component attribute (6.4.3.1).',
                element)

        # Get component
        try:
            component = model.component(component)
        except KeyError:
            raise CellMLParsingError(
                'A component_ref\'s component attribute must reference a'
                ' component in the same model, got "' + component
                + '" (6.4.3.3).', element)

        # Check allowed content
        self._check_allowed_content(element, ['component_ref'], ['component'])

        # Store encapsulation relationships
        # Ignore all other types of relationship
        if parent is not None and 'encapsulation' in relationships:

            # Check for existing parent
            if component.parent() is not None:
                raise CellMLParsingError(
                    'A component can only have a single encapsulation parent:'
                    ' found ' + str(component) + ' with parents '
                    + str(component.parent()) + ' and ' + str(parent)
                    + ' (6.4.3.2).', element)

            # Set parent (won't raise CellMLErrors)
            component.set_parent(parent)

        # Check child component refs
        kids = element.findall(self._join('component_ref'))
        for child in kids:
            self._parse_group_component_ref(
                child, model, relationships, component)

        # No parent? Then (encapsulation and containment) relationships must
        # have at least one child.
        if len(kids) == 0 and parent is None:
            if ('encapsulation' in relationships
                    or 'containment' in relationships):
                raise CellMLParsingError(
                    'The first component_ref in an encapsulation or'
                    ' containment relationship must have at least one child'
                    ' (6.4.3.2).')

    def _parse_group_relationship_ref(self, element, relationships):
        """
        Parses a relationship_ref ``element`` and adds it to the set of
        ``relationships``.

        Named containment relationships will be added as ``containment, name``.
        """
        # Check cmeta id
        self._check_cmeta_id(element)

        # Check relationship is specified
        ns = None
        rel = element.attrib.get('relationship', None)
        if rel is None:
            # Check for relationship attribute in other namespace
            for at in element.attrib:
                ns, at = split(at)
                if at == 'relationship':
                    rel = 'other'
                    break
            if rel is None:
                raise CellMLParsingError(
                    'Relationship_ref must define a relationship attribute'
                    ' (6.4.2.1).', element)

        # Check type (only if in null namespace)
        if ns is None and rel not in ['encapsulation', 'containment']:
            raise CellMLParsingError(
                'Unknown relationship type: "' + rel + '", expecting either'
                ' "encapsulation" or "containment" (6.4.2.2).', element)

        # Check name, if given
        name = element.attrib.get('name', None)
        if name is not None:
            if rel == 'encapsulation':
                raise CellMLParsingError(
                    'Encapsulation relationships may not define a name'
                    ' attribute (6.4.2.4).', element)

            if not myokit.formats.cellml.v1.is_valid_identifier(name):
                raise CellMLParsingError(
                    'Relationship_ref name must be a valid CellML identifier,'
                    ' but found "' + name + '" (6.4.2.3).', element)
            rel += ', ' + name

        # Check uniqueness of relationship (only in null namespace)
        if ns is None and rel in relationships:
            raise CellMLParsingError(
                'Relationship_refs in each group must have a unique pair of'
                ' (relationship, name) attributes (6.4.2.5).', element)

        # Store relationsip
        relationships.add(rel)

        # Check allowed content
        self._check_allowed_content(element, [], ['relationship', 'name'])

    def _parse_math(self, element, component):
        """
        Parses a mathml:math ``element``, adding equations to the variables of
        the given ``component``.
        """

        # Get variables from component
        def variable_factory(name, element):
            try:
                var = component.variable(name)
            except KeyError:
                raise CellMLParsingError(
                    'Variable references in equation must name a variable from'
                    ' the local component (4.4.2.1).', element)

            return myokit.Name(var)

        # Create numbers with units
        attr = self._join('units')

        def number_factory(value, element):
            # Numbers not connected to a cn
            if element is None:
                return myokit.Number(value, myokit.units.dimensionless)

            # Get units attribute
            try:
                units = element.attrib[attr]
            except KeyError:
                raise CellMLParsingError(
                    'Numbers inside MathML must define a cellml:units'
                    ' attribute (4.4.3.1).', element)

            # Find units in component
            try:
                units = component.find_units(units)
                units = units.myokit_unit()
            except myokit.formats.cellml.v1.UnitsError as e:
                warnings.warn(
                    'The units "' + str(units) + '" (referenced inside a'
                    ' MathML equation) are not supported and have been'
                    ' replaced by `dimensionless`. (' + str(e) + ')')
                units = myokit.units.dimensionless
            except myokit.formats.cellml.v1.CellMLError:
                raise CellMLParsingError(
                    'Unknown unit "' + str(units) + '" referenced inside a'
                    ' MathML equation (4.4.3.2).', element)

            # Create and return
            return myokit.Number(value, units)

        # Create parser
        p = myokit.formats.mathml.MathMLParser(
            variable_factory, number_factory, self._free_vars)

        # Iterate over applies, allowing for applies nested inside <semantics>
        # elements
        todo = collections.deque([x for x in element])
        while todo:
            child = todo.popleft()

            # Check each child is in MathML namespace
            ns, el = split(child.tag)
            if ns != cellml.NS_MATHML:
                raise CellMLParsingError(
                    'The contents of a mathml:math element must be in the'
                    ' mathml namespace, found "' + str(child.tag) + '" inside '
                    + str(component) + '.', child)

            # Ignore annotations
            if el in ['annotation', 'annotation-xml']:
                continue

            # Look inside of semantics elements
            if el == 'semantics':
                todo.extendleft(reversed(child))
                continue

            # If it isn't these it must be an apply
            if el != 'apply':
                raise CellMLParsingError(
                    'Unexpected contents in mathml:math. Expecting'
                    ' mathml:apply but found mathml:' + el + ' inside maths'
                    ' for ' + str(component) + '.', child)

            # Parse
            eq = p.parse(child)
            if not isinstance(eq, myokit.Equal):
                raise CellMLParsingError(
                    'Unexpected element in MathML, expecting a list of'
                    ' equations, got ' + self._tag(child) + '.', child)
            lhs, rhs = eq

            # Check lhs
            if not isinstance(lhs, myokit.LhsExpression):
                raise CellMLParsingError(
                    'Invalid expression found on the left-hand side of an'
                    ' equation: ' + self._dae_message, child)

            # Promote derivative, check non-derivatives have no initial value
            var = lhs.var()
            if isinstance(lhs, myokit.Derivative):
                var.set_is_state(True)
            elif var.initial_value() is not None:
                raise CellMLParsingError(
                    'Initial value and a defining equation found for'
                    ' non-state ' + str(var) + ': ' + self._dae_message,
                    child)

            # Check for double rhs
            if var.rhs() is not None:
                raise CellMLParsingError(
                    'Two defining equations found for ' + str(var) + ': '
                    + self._dae_message,
                    child)

            # Set rhs
            try:
                lhs.var().set_rhs(rhs)
            except myokit.formats.cellml.v1.CellMLError as e:
                raise CellMLParsingError(str(e), child)

    def _parse_model(self, element):
        """
        Parses a CellML model element.
        """
        # Handle document-level validation here.

        # Check namespace
        ns, el = split(element.tag)
        if ns == cellml.NS_CELLML_1_0:
            version = '1.0'
        elif ns == cellml.NS_CELLML_1_1:
            version = '1.1'
        else:
            raise CellMLParsingError(
                'Root node must be in CellML 1.0 or 1.1 namespace.', element)

        # Store namespace
        self._ns = ns

        # Check root element is a model
        if el != 'model':
            raise CellMLParsingError(
                'Root node must be a CellML model element.', element)

        # Check name is present
        try:
            name = element.attrib['name']
        except KeyError:
            raise CellMLParsingError(
                'Model element must have a name attribute (3.4.1.1).', element)

        # Create model (validates name)
        try:
            model = myokit.formats.cellml.v1.Model(name, version)

            # Check cmeta id
            cmeta_id = self._check_cmeta_id(element)
            if cmeta_id:
                model.set_cmeta_id(cmeta_id)

        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e), element)

        # Check for imports
        im = element.find(self._join('import'))
        if im is not None:
            if version == '1.1':
                raise CellMLParsingError('Imports are not supported.', im)
            else:
                raise CellMLParsingError(
                    'Imports are not allowed in CellML 1.0.', im)

        # Check allowed content
        self._check_allowed_content(
            element,
            ['component', 'connection', 'group', 'units'],
            ['name'],
            name,
        )

        # Create model units
        for child in self._sort_units(element):
            self._parse_units(child, model)

        # Create components
        components = element.findall(self._join('component'))
        for child in components:
            self._parse_component(child, model)

        # Create encapsulation hierarchy
        for child in element.findall(self._join('group')):
            self._parse_group(child, model)

        # Add connections
        connected = set()
        for child in element.findall(self._join('connection')):
            self._parse_connection(child, model, connected)

        # Add equations
        for child in components:
            component = model.component(child.attrib['name'])
            for math in child.findall(self._join('math', cellml.NS_MATHML)):
                self._parse_math(math, component)

        # Check number of free variables
        self._free_vars = set(
            [x.var().value_source() for x in self._free_vars])
        if len(self._free_vars) > 1:
            raise CellMLParsingError(
                'Models that take derivatives with respect to more than one'
                ' variable are not supported.')

        elif len(self._free_vars) == 1:
            # Set free variable
            model.set_free_variable(self._free_vars.pop())

        # Read any rdf annotations
        rdf_tag = self._join('RDF', cellml.NS_RDF)
        for child in element.findall(rdf_tag):
            self._parse_rdf(child, model)

        # Read any documentation
        doc_tag = self._join('documentation', cellml.NS_TMP_DOC)
        for child in element.findall(doc_tag):
            self._parse_documentation(child, model)

        # Perform final validation and return
        model.validate()

        return model

    def _parse_rdf(self, element, model):
        """
        Parses an RDF tag, looks for oxmeta annotations.
        """
        # rdf:Description tag with rdf:about attribute
        dtag = self._join('Description', cellml.NS_RDF)
        aatt = self._join('about', cellml.NS_RDF)

        # bqbiol:is tag with rdf:resource attribute
        itag = self._join('is', cellml.NS_BQBIOL)
        ratt = self._join('resource', cellml.NS_RDF)

        # Oxmeta namespace
        ox = cellml.NS_OXMETA

        # Look for descriptions
        for child in element.findall(dtag):

            # Get variable cmeta id
            about = child.attrib.get(aatt, None)
            if about is None:
                continue

            # Strip leading '#'
            about = about[1:]

            # Look for variable with that cmeta id
            try:
                variable = model.element_with_cmeta_id(about)
            except KeyError:
                continue
            if not isinstance(variable, myokit.formats.cellml.v1.Variable):
                continue

            # Get bqbiol:is element
            eis = child.find(itag)
            if eis is None:
                continue

            # Get resource
            resource = eis.attrib.get(ratt, None)
            if resource is None:
                continue

            # Strip oxmeta from annotation and store
            if not resource.startswith(ox):
                continue
            annotation = resource[len(ox):]
            variable.meta['oxmeta'] = annotation

    def _parse_unit(self, element, owner):
        """
        Parses a unit ``element`` and returns a :class:`myokit.Unit`, using the
        :class:`Model` or :class:`Component` ``owner`` is used to look up
        units definitions.
        """

        # Get units attribute (already checked)
        units = element.attrib['units']

        # Check cmeta id
        self._check_cmeta_id(element)

        # Get prefix, exponent, and multiplier
        prefix = element.attrib.get('prefix', 0)
        exponent = element.attrib.get('exponent', 1)
        multiplier = element.attrib.get('multiplier', 1)

        # Check for offset
        x = element.attrib.get('offset', 0)
        try:
            x = float(x)
        except ValueError:
            raise CellMLParsingError(
                'Unit offset must be a real number (5.4.2.6).', element)
        if x != 0:
            raise myokit.formats.cellml.v1.UnsupportedUnitOffsetError

        # Check allowed content
        self._check_allowed_content(
            element,
            [],
            ['units', 'prefix', 'exponent', 'multiplier', 'offset'],
        )

        # Create and return
        return myokit.formats.cellml.v1.Units.parse_unit_row(
            units, prefix, exponent, multiplier, owner)

    def _parse_units(self, element, owner):
        """
        Parses a units ``element``, adding the resulting units definition to
        the :class:`Model` or :class:`Component` ``owner``.
        """

        # Name has already been checked at this point
        name = element.attrib['name']

        # Check if this is a base unit (not supported)
        base = element.attrib.get('base_units', 'no')
        if base not in ('yes', 'no'):
            raise CellMLParsingError(
                'Base units attribute must be either "yes" or "no".', element)

        # Check cmeta id
        self._check_cmeta_id(element)

        # Check allowed content
        self._check_allowed_content(
            element, ['unit'], ['name', 'base_units'], name)

        # Check the units definition has children
        children = element.findall(self._join('unit'))
        if not children:
            if base == 'yes':
                warnings.warn(
                    'Unable to parse definition for units "' + str(name) + '",'
                    ' using `dimensionless` instead. (Defining new base units'
                    ' is not supported.)')
            else:
                raise CellMLParsingError(
                    'Units element with base_units="no" must contain at least'
                    ' one child unit element.')

        # Parse content
        myokit_unit = myokit.units.dimensionless
        try:
            for child in children:
                myokit_unit *= self._parse_unit(child, owner)
        except myokit.formats.cellml.v1.UnitsError as e:
            warnings.warn(
                'Unable to parse definition for units "' + str(name) + '",'
                ' using `dimensionless` instead. (' + str(e) + ')')
            myokit_unit = myokit.units.dimensionless

        # Add units to owner
        try:
            owner.add_units(name, myokit_unit)
        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e), element)

    def _parse_variable(self, element, component):
        """
        Parses a variable ``element`` and adds a variable to the given
        ``component``.
        """
        # Check name is present
        try:
            name = element.attrib['name']
        except KeyError:
            raise CellMLParsingError(
                'Variable element must have a name attribute (3.4.3.1).',
                element)

        # Check units are present
        try:
            units = element.attrib['units']
        except KeyError:
            raise CellMLParsingError(
                'Variable element must have a units attribute (3.4.3.1).',
                element)

        # Get interface
        pub = element.attrib.get('public_interface', 'none')
        pri = element.attrib.get('private_interface', 'none')

        # Create variable (validates name, units, and name uniqueness)
        try:
            try:
                variable = component.add_variable(name, units, pub, pri)
            except myokit.formats.cellml.v1.UnitsError as e:
                warnings.warn(
                    'The units "' + str(units) + '" (assigned to a variable)'
                    ' are not supported and have been replaced by'
                    ' `dimensionless`. (' + str(e) + ')')
                units = 'dimensionless'
                variable = component.add_variable(name, units, pub, pri)

            # Check cmeta id
            cmeta_id = self._check_cmeta_id(element)
            if cmeta_id:
                variable.set_cmeta_id(cmeta_id)

            # Check allowed content
            attr = [
                'initial_value',
                'name',
                'public_interface',
                'private_interface',
                'units',
            ]
            self._check_allowed_content(element, [], attr, name)

            # Set initial value
            variable.set_initial_value(
                element.attrib.get('initial_value', None))

        except myokit.formats.cellml.v1.CellMLError as e:
            raise CellMLParsingError(str(e), element)

    def _sort_units(self, element, model=None):
        """
        Returns all units elements in ``element``, sorted in an order that
        makes them resolvable.

        Arguments

        ``element``
            A model or component element
        ``model``
            If the owner is a component, a
            :class:`myokit.formats.cellml.v1.Model` can be passed in to look up
            any model units.

        """
        # Create a list of resolved and a dict mapping unresolved unit names to
        # their unknown dependencies
        resolved = set()
        unresolved = {}

        # Add SI units to the list of resolved units
        si_units = set(myokit.formats.cellml.v1.Units.si_unit_names())
        resolved.update(si_units)

        # If this is a component, add any model units to the list
        if model is not None:
            for units in model.units():
                resolved.add(units.name())

        # Mapping of local units to their elementtree elements
        local_units = {}

        # Populate the set of unresolved units.
        for units in element.findall(self._join('units')):
            # Get name, complain if it doesn't exist
            try:
                name = units.attrib['name']
            except KeyError:
                raise CellMLParsingError(
                    'Units element must have a name attribute (5.4.1.1).',
                    element)

            # Check doesn't shadow an si unit
            if name in si_units:
                raise CellMLParsingError(
                    'Units name "' + name + '" overlaps with a predefined name'
                    ' in ' + self._tag(element) + ' (5.4.1.2).', element)

            # Check for duplicates
            if name in local_units:
                raise CellMLParsingError(
                    'Duplicate units definition "' + name + '" in '
                    + self._tag(element) + '.', element)
            local_units[name] = units

            # Component units can shadow model units, so if known units are
            # found they will become 'unresolved' at this point.
            if name in resolved:
                resolved.remove(name)

            # Determine dependencies
            deps = set()
            for unit in units.findall(self._join('unit')):
                try:
                    dep = unit.attrib['units']
                except KeyError:
                    raise CellMLParsingError(
                        'Unit elements must have a units attribute (5.4.2.1).',
                        element)
                deps.add(dep)
            unresolved[name] = deps

        # Remove resolved variables from deps
        for name, deps in unresolved.items():
            deps.difference_update(resolved)

        # Sort
        ordered = []
        while unresolved:
            fresh = set()
            for name, deps in unresolved.items():
                if not deps:
                    fresh.add(name)
                    ordered.append(local_units[name])
            if fresh:
                for name in fresh:
                    resolved.add(name)
                    del unresolved[name]
                for name, deps in unresolved.items():
                    deps.difference_update(fresh)
            else:
                raise CellMLParsingError(
                    'Unable to resolve network of units in '
                    + self._tag(element) + ' (5.4.2.2).', element)
        return ordered

    def _tag(self, element, name=None):
        """
        Returns an element's name, but changes the syntax from ``{...}tag`` to
        ``cellml:tag`` for CellML elements.

        If the element is in a CellML namespace and an optional ``name``
        attribute is given, this is added to the returned output using xpath
        syntax, e.g. ``cellml:model[@name="MyModel"]``.

        Can also be used for attributes.
        """
        ns, el = split(element.tag)
        tag = self._item(ns, el)
        if ns == self._ns and name is not None:
            tag += '[@name="' + name + '"]'
        return tag




#!/usr/bin/env python3
#
# Tests the importers for various formats
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats as formats

from myokit.tests import DIR_FORMATS


class ImporterTest(unittest.TestCase):
    """ Test shared importer functionality. """

    def test_importer_interface(self):
        # Test listing and creating importers.
        ims = myokit.formats.importers()
        self.assertTrue(len(ims) > 0)
        for i in ims:
            self.assertIsInstance(i, str)
            i = myokit.formats.importer(i)
            self.assertTrue(isinstance(i, myokit.formats.Importer))

    def test_unknown(self):
        # Test requesting an unknown importer.
        # Test fetching using importer method
        self.assertRaisesRegex(
            KeyError, 'Importer not found', myokit.formats.importer, 'blip')


class AxonTest(unittest.TestCase):
    """ Partially tests Axon formats importing. """

    def test_capability_reporting(self):
        # Test if the right capabilities are reported.
        i = formats.importer('abf')
        self.assertFalse(i.supports_component())
        self.assertFalse(i.supports_model())
        self.assertTrue(i.supports_protocol())

    def test_protocol(self):
        i = formats.importer('abf')
        self.assertTrue(i.supports_protocol())
        i.protocol(os.path.join(DIR_FORMATS, 'abf-v1.abf'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests the DataBlock1d and DataBlock2d classes
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import numpy as np

import myokit

from myokit.tests import (
    TemporaryDirectory,
    DIR_DATA,
    DIR_IO,
    TestReporter,
    CancellingReporter,
    WarningCollector,
)


class DataBlock1dTest(unittest.TestCase):
    """
    Tests the DataBlock1d
    """
    def test_combined(self):
        # Runs a combined test of:
        #
        # - DataLog to Block conversion
        # - Access to fields in block
        # - Saving block to binary file
        # - Loading block from binary file
        #

        # Create simulation log with 1d data
        log = myokit.DataLog()
        log.set_time_key('time')
        t = np.linspace(0, 1, 20)
        log['time'] = t
        log['0.x'] = np.sin(t)
        log['1.x'] = np.cos(t)
        log['2.x'] = np.tan(t)

        # Convert to datablock
        b = log.block1d()

        # Check block contents
        self.assertTrue(np.all(b.time() == t))
        self.assertFalse(b.time() is t)
        self.assertEqual(b.len0d(), 0)
        self.assertEqual(b.len1d(), 1)
        x = b.get1d('x')
        self.assertTrue(np.all(x[:, 0] == log['0.x']))
        self.assertTrue(np.all(x[:, 1] == log['1.x']))
        self.assertTrue(np.all(x[:, 2] == log['2.x']))

        # Make bigger log, try again
        log['pace'] = np.ones(t.shape) + t**2
        log['0.y'] = np.sqrt(t)
        log['1.y'] = 1 + np.sqrt(t)
        log['2.y'] = 2 + np.sqrt(t)

        # Convert to datablock
        b = log.block1d()

        # Check block contents
        self.assertTrue(np.all(b.time() == t))
        self.assertFalse(b.time() is t)
        self.assertEqual(b.len0d(), 1)
        self.assertEqual(b.len1d(), 2)
        self.assertTrue(np.all(b.get0d('pace') == log['pace']))
        self.assertFalse(b.get0d('pace') is log['pace'])
        x = b.get1d('x')
        self.assertTrue(np.all(x[:, 0] == log['0.x']))
        self.assertTrue(np.all(x[:, 1] == log['1.x']))
        self.assertTrue(np.all(x[:, 2] == log['2.x']))
        y = b.get1d('y')
        self.assertTrue(np.all(y[:, 0] == log['0.y']))
        self.assertTrue(np.all(y[:, 1] == log['1.y']))
        self.assertTrue(np.all(y[:, 2] == log['2.y']))

        # Test reading and writing
        with TemporaryDirectory() as d:
            fname = d.path('block1d.zip')
            b.save(fname)
            c = myokit.DataBlock1d.load(fname)
            # Test block contents
            self.assertTrue(np.all(b.time() == c.time()))
            self.assertFalse(b.time() is c.time())
            self.assertEqual(c.len0d(), 1)
            self.assertEqual(c.len1d(), 2)
            self.assertTrue(np.all(b.get0d('pace') == c.get0d('pace')))
            self.assertFalse(b.get0d('pace') is c.get0d('pace'))
            xb = b.get1d('x')
            xc = c.get1d('x')
            self.assertTrue(np.all(xb == xc))
            self.assertFalse(xb is xc)
            yb = b.get1d('y')
            yc = c.get1d('y')
            self.assertTrue(np.all(yb == yc))
            self.assertFalse(yb is yc)

    def test_bad_constructor(self):
        # Test bad arguments to a DataBlock1d raise ValueErrors.

        # Test valid constructor
        w = 2
        time = [1, 2, 3]
        myokit.DataBlock1d(w, time)

        # Test w < 1
        self.assertRaises(ValueError, myokit.DataBlock1d, 0, time)

        # Test time matrix
        self.assertRaises(ValueError, myokit.DataBlock1d, w, [[1, 2, 3]])

        # Decreasing times
        self.assertRaises(ValueError, myokit.DataBlock1d, w, [3, 2, 1])

    def test_block2d(self):
        # Test conversion to 2d.

        w = 2
        time = [1, 2, 3]
        pace = [0, 0, 2]
        x = np.array([[1, 1], [2, 2], [3, 3]])
        b1 = myokit.DataBlock1d(w, time)
        b1.set0d('pace', pace)
        b1.set1d('x', x)

        b2 = b1.block2d()
        self.assertTrue(np.all(b2.get0d('pace') == pace))
        self.assertTrue(np.all(b2.get2d('x') == x.reshape((3, 1, 2))))

    def test_cv(self):
        # Test the CV method.

        b = os.path.join(DIR_DATA, 'cv1d.zip')
        b = myokit.DataBlock1d.load(b)
        self.assertAlmostEqual(b.cv('membrane.V'), 5.95272837350686004e+01)

        # Invalid border argument
        # Negative
        self.assertRaises(ValueError, b.cv, 'membrane.V', border=-1)
        # Too large
        self.assertRaises(ValueError, b.cv, 'membrane.V', border=1000)

        # No upstroke
        time = np.linspace(0, 1, 100)
        down = np.zeros(100) - 85

        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = time
        d['x', 0] = down
        d['x', 1] = down
        d['x', 2] = down
        d['x', 3] = down
        d['x', 4] = down
        b = myokit.DataBlock1d.from_log(d)
        self.assertEqual(b.cv('x'), 0)

        # No propagation: Single-cell stimulus
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = time
        d['x', 0] = np.array(down, copy=True)
        d['x', 1] = np.array(down, copy=True)
        d['x', 2] = np.array(down, copy=True)
        d['x', 3] = np.array(down, copy=True)
        d['x', 4] = np.array(down, copy=True)
        d['x', 1][10:] = 40
        b = myokit.DataBlock1d.from_log(d)
        self.assertEqual(b.cv('x'), 0)

        # No propagation: Multi-cell stimulus
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = time
        d['x', 0] = np.array(down, copy=True)
        d['x', 1] = np.array(down, copy=True)
        d['x', 2] = np.array(down, copy=True)
        d['x', 3] = np.array(down, copy=True)
        d['x', 4] = np.array(down, copy=True)
        d['x', 1][10:] = 40
        d['x', 2][10:] = 40
        d['x', 3][10:] = 40
        b = myokit.DataBlock1d.from_log(d)
        self.assertEqual(b.cv('x'), 0)

    def test_from_log(self):
        # Test some edge cases of `DataBlock1d.from_log`.

        # Test valid construction
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = [1, 2, 3]
        d['x', 0] = [0, 0, 0]
        d['x', 1] = [1, 1, 2]
        myokit.DataBlock1d.from_log(d)

        # Deprecated alias
        with WarningCollector() as wc:
            myokit.DataBlock1d.from_DataLog(d)
        self.assertIn('deprecated', wc.text())

        # No time key
        d = myokit.DataLog()
        d['time'] = [1, 2, 3]
        d['x', 0] = [0, 0, 0]
        d['x', 1] = [1, 1, 2]
        self.assertRaises(ValueError, myokit.DataBlock1d.from_log, d)

        # Multi-dimensional time key
        d.set_time_key('0.x')
        self.assertRaises(ValueError, myokit.DataBlock1d.from_log, d)

        # 2-dimensional stuff
        d.set_time_key('time')
        myokit.DataBlock1d.from_log(d)
        d['y', 0, 0] = [10, 10, 10]
        self.assertRaises(ValueError, myokit.DataBlock1d.from_log, d)

        # Mismatched dimensions
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = [1, 2, 3]
        d['x', 0] = [0, 0, 0]
        d['x', 1] = [1, 1, 2]
        d['y', 0] = [2, 0, 0]
        d['y', 1] = [3, 1, 2]
        d['y', 2] = [0, 4, 5]
        self.assertRaises(ValueError, myokit.DataBlock1d.from_log, d)

    def test_grids(self):
        # Test conversion of the DataBlock1d to a 2d grid.

        # Create simulation log with 1d data
        log = myokit.DataLog()
        log.set_time_key('time')
        n = 5   # Points in time
        w = 8   # Points in space
        t = np.arange(n)
        log['time'] = t
        for i in range(w):
            log['x', i] = i * t
            log['y', i] = i * t * 10

        # Create datablock
        b = log.block1d()

        # Check dimensions
        self.assertEqual(b.len0d(), 0)
        self.assertEqual(b.len1d(), 2)
        self.assertEqual(b.shape(), (n, w))
        self.assertEqual(n, len(b.time()))
        self.assertEqual(b.get1d('x').shape, (n, w))
        self.assertEqual(b.get1d('y').shape, (n, w))

        # Convert to grid for matplotlib plotting
        x, y, z = b.grid('x', transpose=True)
        self.assertEqual(x.shape, (w + 1, n + 1))
        self.assertEqual(y.shape, (w + 1, n + 1))
        self.assertEqual(z.shape, (w, n))
        xx, yy, zz = b.grid('x', transpose=False)
        self.assertEqual(xx.shape, (n + 1, w + 1))
        self.assertFalse(np.all(xx.transpose() == x))
        self.assertEqual(yy.shape, (n + 1, w + 1))
        self.assertFalse(np.all(yy.transpose() == y))
        self.assertEqual(zz.shape, (n, w))
        self.assertTrue(np.all(zz.transpose() == z))

        # Convert to grid for image-like plotting
        x, y, z = b.image_grid('x', transpose=True)
        self.assertEqual(x.shape, (w, n))
        self.assertEqual(y.shape, (w, n))
        self.assertEqual(z.shape, (w, n))
        self.assertTrue(
            np.all(x == np.array([
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4],
                [0, 1, 2, 3, 4]])))
        self.assertTrue(
            np.all(y == np.array([
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [2, 2, 2, 2, 2],
                [3, 3, 3, 3, 3],
                [4, 4, 4, 4, 4],
                [5, 5, 5, 5, 5],
                [6, 6, 6, 6, 6],
                [7, 7, 7, 7, 7]])))
        self.assertTrue(
            np.all(z == np.array([
                [0, 0, 0, 0, 0],
                [0, 1, 2, 3, 4],
                [0, 2, 4, 6, 8],
                [0, 3, 6, 9, 12],
                [0, 4, 8, 12, 16],
                [0, 5, 10, 15, 20],
                [0, 6, 12, 18, 24],
                [0, 7, 14, 21, 28]])))
        xx, yy, zz = b.image_grid('x', transpose=False)
        self.assertEqual(xx.shape, (n, w))
        self.assertEqual(yy.shape, (n, w))
        self.assertEqual(zz.shape, (n, w))
        self.assertFalse(np.all(xx == x.transpose()))
        self.assertTrue(np.all(xx == np.array([
            [0, 1, 2, 3, 4, 5, 6, 7],
            [0, 1, 2, 3, 4, 5, 6, 7],
            [0, 1, 2, 3, 4, 5, 6, 7],
            [0, 1, 2, 3, 4, 5, 6, 7],
            [0, 1, 2, 3, 4, 5, 6, 7]])))
        self.assertFalse(np.all(yy == y.transpose()))
        self.assertTrue(np.all(yy == np.array([
            [0, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2, 2, 2, 2],
            [3, 3, 3, 3, 3, 3, 3, 3],
            [4, 4, 4, 4, 4, 4, 4, 4]])))
        self.assertTrue(np.all(zz == z.transpose()))

        # Compare grid and image_grid
        x, y, z = b.grid('x', transpose=True)
        xx, yy, zz = b.image_grid('x', transpose=True)
        self.assertTrue(np.all(x[:-1, :-1] == xx))
        self.assertTrue(np.all(y[:-1, :-1] == yy))
        self.assertTrue(np.all(z == zz))

    def test_keys(self):
        # Test the keys0d() and keys1d() methods.

        w = 2
        time = [1, 2, 3]
        b1 = myokit.DataBlock1d(w, time)

        # Test keys0d
        self.assertEqual(len(list(b1.keys0d())), 0)
        pace = np.array([0, 0, 2])
        b1.set0d('pace', pace)
        self.assertEqual(len(list(b1.keys0d())), 1)
        self.assertIn('pace', b1.keys0d())
        b1.set0d('poos', pace)
        self.assertEqual(len(list(b1.keys0d())), 2)
        self.assertIn('poos', b1.keys0d())

        # Test keys1d
        self.assertEqual(len(list(b1.keys1d())), 0)
        x = np.array([[1, 1], [2, 2], [3, 3]])
        b1.set1d('x', x)
        self.assertEqual(len(list(b1.keys1d())), 1)
        y = np.array([[2, 1], [3, 2], [4, 3]])
        b1.set1d('y', y)
        self.assertEqual(len(list(b1.keys1d())), 2)

    def test_load_bad_file(self):
        # Test loading errors.

        # Not enough files
        path = os.path.join(DIR_IO, 'bad1d-1-not-enough-files.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Not enough files',
            myokit.DataBlock1d.load, path)

        # No header file
        path = os.path.join(DIR_IO, 'bad1d-2-no-header.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Header not found',
            myokit.DataBlock1d.load, path)

        # No data file
        path = os.path.join(DIR_IO, 'bad1d-3-no-data.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Data not found',
            myokit.DataBlock1d.load, path)

        # Not a zip
        path = os.path.join(DIR_IO, 'bad1d-4-not-a-zip.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Bad zip',
            myokit.DataBlock1d.load, path)

        # Unknown data type in data
        path = os.path.join(DIR_IO, 'bad1d-5-bad-data-type.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Unrecognized data type',
            myokit.DataBlock1d.load, path)

        # Not enough data: detected at time level
        path = os.path.join(DIR_IO, 'bad1d-6-time-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock1d.load, path)

        # Not enoug data: detected at 0d level
        path = os.path.join(DIR_IO, 'bad1d-7-0d-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock1d.load, path)

        # Not enough data: detected at 1d level
        path = os.path.join(DIR_IO, 'bad1d-8-1d-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock1d.load, path)

        # Test progress reporter
        path = os.path.join(DIR_DATA, 'cv1d.zip')
        p = TestReporter()
        self.assertFalse(p.entered)
        self.assertFalse(p.exited)
        self.assertFalse(p.updated)
        b = myokit.DataBlock1d.load(path, p)
        self.assertIsNotNone(b)
        self.assertTrue(p.entered)
        self.assertTrue(p.exited)
        self.assertTrue(p.updated)

        # Test cancelling in progress reporter
        p = CancellingReporter()
        b = myokit.DataBlock1d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(1)
        b = myokit.DataBlock1d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(2)
        b = myokit.DataBlock1d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(3)
        b = myokit.DataBlock1d.load(path, p)
        self.assertIsNone(b)

    def test_remove0d(self):
        # Test remove0d().

        # Add new 0d field and remove again
        b = myokit.DataBlock1d(2, [1, 2, 3])
        self.assertEqual(b.len0d(), 0)
        b.set0d('pace', np.array([0, 0, 2]))
        self.assertEqual(b.len0d(), 1)
        b.remove0d('pace')
        self.assertEqual(b.len0d(), 0)

        # Time can't be removed
        self.assertRaises(KeyError, b.remove0d, 'time')

    def test_remove1d(self):
        # Test remove1d().

        # Add new 0d field and remove again
        b = myokit.DataBlock1d(2, [1, 2, 3])
        self.assertEqual(b.len1d(), 0)
        b.set1d('z', np.array([[0, 0], [2, 3], [2, 1]]))
        self.assertEqual(b.len1d(), 1)
        b.remove1d('z')
        self.assertEqual(b.len1d(), 0)

    def test_set0d(self):
        # Test set0d().

        w = 2
        time = [1, 2, 3]
        b1 = myokit.DataBlock1d(w, time)

        # Test valid call
        pace = np.array([0, 0, 2])
        b1.set0d('pace', pace)
        self.assertTrue(np.all(b1.get0d('pace') == pace))

        # Test copying
        b1.set0d('pace', pace, copy=False)
        self.assertTrue(np.all(b1.get0d('pace') == pace))
        self.assertIs(b1.get0d('pace'), pace)
        b1.set0d('pace', pace, copy=True)
        self.assertTrue(np.all(b1.get0d('pace') == pace))
        self.assertIsNot(b1.get0d('pace'), pace)

        # Test bad calls
        self.assertRaises(ValueError, b1.set0d, '', pace)
        self.assertRaises(ValueError, b1.set0d, 'pace', [1, 2, 3, 4])
        self.assertRaises(ValueError, b1.set0d, 'pace', np.array([1, 2, 3, 4]))

    def test_set1d(self):
        # Test set1d().

        w = 2
        time = [1, 2, 3]
        b1 = myokit.DataBlock1d(w, time)

        # Test valid call
        x = np.array([[1, 1], [2, 2], [3, 3]])
        b1.set1d('x', x)
        self.assertTrue(np.all(b1.get1d('x') == x))

        # Test copying
        b1.set1d('x', x, copy=False)
        self.assertTrue(np.all(b1.get1d('x') == x))
        self.assertIs(b1.get1d('x'), x)
        b1.set1d('x', x, copy=True)
        self.assertTrue(np.all(b1.get1d('x') == x))
        self.assertIsNot(b1.get1d('x'), x)

        # Test bad calls
        self.assertRaises(ValueError, b1.set1d, '', x)
        x = np.array([[1, 1], [2, 2], [3, 3], [4, 4]])
        self.assertRaises(ValueError, b1.set1d, 'x', x)

    def test_to_log(self):
        # Test the `DataBlock1d.to_log()` method.

        # Create test block
        b = myokit.DataBlock1d(3, [1, 2, 3, 4])
        b.set0d('pace', [0, 1, 0, 0])
        b.set1d('voltage', [[2, 1, 0], [1, 2, 1], [0, 1, 2], [0, 0, 1]])

        # Convert and inspect
        d = b.to_log()
        self.assertIn('time', d)
        self.assertIn('pace', d)
        self.assertIn('0.voltage', d)
        self.assertIn('1.voltage', d)
        self.assertIn('2.voltage', d)
        self.assertEqual(len(d), 5)
        self.assertEqual(list(d.time()), [1, 2, 3, 4])
        self.assertEqual(list(d['pace']), [0, 1, 0, 0])
        self.assertEqual(list(d['0.voltage']), [2, 1, 0, 0])
        self.assertEqual(list(d['1.voltage']), [1, 2, 1, 0])
        self.assertEqual(list(d['2.voltage']), [0, 1, 2, 1])

    def test_trace(self):
        w = 2
        time = [1, 2, 3]
        b = myokit.DataBlock1d(w, time)
        x = np.array([[1, 4], [2, 5], [3, 6]])
        b.set1d('x', x)
        self.assertTrue(np.all(b.trace('x', 0) == [1, 2, 3]))
        self.assertTrue(np.all(b.trace('x', 1) == [4, 5, 6]))


class DataBlock2dTest(unittest.TestCase):
    """
    Tests `DataBlock2d`.
    """

    def test_bad_constructor(self):
        # Test bad arguments to a DataBlock2d raise ValueErrors.

        # Test valid constructor
        w = 2
        h = 3
        time = [1, 2, 3]
        myokit.DataBlock2d(w, h, time)

        # Test w < 1, h < 1
        self.assertRaises(ValueError, myokit.DataBlock2d, 0, h, time)
        self.assertRaises(ValueError, myokit.DataBlock2d, w, 0, time)

        # Test time matrix
        self.assertRaises(ValueError, myokit.DataBlock2d, w, h, [[1, 2, 3]])

        # Decreasing times
        self.assertRaises(ValueError, myokit.DataBlock2d, w, h, [3, 2, 1])

    def test_colors(self):
        # Test conversion to colors using different color maps.

        w, h = 2, 3
        time = [1, 2, 3]
        b = myokit.DataBlock2d(w, h, time)
        x = np.array([  # Each 3 by 2 array is a point in time
            [[0, 1],
             [2, 3],
             [4, 5]],

            [[5, 4],
             [3, 2],
             [1, 0]],

            [[0, 0],
             [0, 0],
             [0, 0]],
        ])
        b.set2d('x', x)

        # Red colormap
        t0 = np.array([  # Deepest array is a pixel
            [[255, 255, 255], [255, 204, 204]],
            [[255, 153, 153], [255, 102, 102]],
            [[255, 50, 50], [255, 0, 0]],
        ])
        t1 = np.array([
            [[255, 0, 0], [255, 50, 50]],
            [[255, 102, 102], [255, 153, 153]],
            [[255, 204, 204], [255, 255, 255]],
        ])
        t2 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        c = b.colors('x', colormap='red')
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

        # Red with normalization
        t0 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 127, 127]],
            [[255, 0, 0], [255, 0, 0]],
        ])
        t1 = np.array([
            [[255, 0, 0], [255, 0, 0]],
            [[255, 127, 127], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        t2 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        c = b.colors('x', colormap='red', lower=2, upper=4)
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

        # Red with extreme normalization
        t2 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        c = b.colors('x', colormap='red', lower=4, upper=4)
        self.assertTrue(np.all(c[0] == t2))
        self.assertTrue(np.all(c[1] == t2))
        self.assertTrue(np.all(c[2] == t2))

        # Green
        t0 = np.array([
            [[255, 255, 255], [204, 255, 204]],
            [[153, 255, 153], [102, 255, 102]],
            [[50, 255, 50], [0, 255, 0]],
        ])
        t1 = np.array([
            [[0, 255, 0], [50, 255, 50]],
            [[102, 255, 102], [153, 255, 153]],
            [[204, 255, 204], [255, 255, 255]],
        ])
        t2 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        c = b.colors('x', colormap='green')
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

        # Blue colormap
        t0 = np.array([  # Deepest array is a pixel
            [[255, 255, 255], [204, 204, 255]],
            [[153, 153, 255], [102, 102, 255]],
            [[50, 50, 255], [0, 0, 255]],
        ])
        t1 = np.array([
            [[0, 0, 255], [50, 50, 255]],
            [[102, 102, 255], [153, 153, 255]],
            [[204, 204, 255], [255, 255, 255]],
        ])
        t2 = np.array([
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
            [[255, 255, 255], [255, 255, 255]],
        ])
        c = b.colors('x', colormap='blue')
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

        # Rainbow/traditional colormap
        t0 = np.array([  # Deepest array is a pixel
            [[153, 61, 143], [49, 35, 212]],
            [[4, 235, 249], [4, 250, 14]],
            [[209, 213, 35], [153, 61, 61]],
        ])
        t1 = np.array([
            [[153, 61, 61], [209, 213, 35]],
            [[4, 250, 14], [4, 235, 249]],
            [[49, 35, 212], [153, 61, 143]],
        ])
        t2 = np.array([
            [[153, 61, 143], [153, 61, 143]],
            [[153, 61, 143], [153, 61, 143]],
            [[153, 61, 143], [153, 61, 143]],
        ])
        c = b.colors('x', colormap='traditional')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

        # Gray colormap
        t0 = np.array([  # Deepest array is a pixel
            [[0] * 3, [51] * 3],
            [[102] * 3, [153] * 3],
            [[204] * 3, [255] * 3],
        ])
        t1 = np.array([
            [[255] * 3, [204] * 3],
            [[153] * 3, [102] * 3],
            [[51] * 3, [0] * 3],
        ])
        t2 = np.array([
            [[0] * 3, [0] * 3],
            [[0] * 3, [0] * 3],
            [[0] * 3, [0] * 3],
        ])
        c = b.colors('x', colormap='gray')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

        # Cividis colormap
        t0 = np.array([  # Deepest array is a pixel
            [[0, 32, 76], [49, 68, 107]],
            [[102, 104, 112], [150, 143, 119]],
            [[203, 186, 104], [255, 233, 69]],
        ])
        t1 = np.array([
            [[255, 233, 69], [203, 186, 104]],
            [[150, 143, 119], [102, 104, 112]],
            [[49, 68, 107], [0, 32, 76]],
        ])
        t2 = np.array([
            [[0, 32, 76], [0, 32, 76]],
            [[0, 32, 76], [0, 32, 76]],
            [[0, 32, 76], [0, 32, 76]],
        ])
        c = b.colors('x', colormap='cividis')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

        # Inferno colormap
        t0 = np.array([  # Deepest array is a pixel
            [[0, 0, 4], [66, 10, 104]],
            [[148, 38, 104], [223, 82, 57]],
            [[253, 167, 12], [253, 255, 165]],
        ])
        t1 = np.array([
            [[253, 255, 165], [253, 167, 12]],
            [[223, 82, 57], [148, 38, 104]],
            [[66, 10, 104], [0, 0, 4]],
        ])
        t2 = np.array([
            [[0, 0, 4], [0, 0, 4]],
            [[0, 0, 4], [0, 0, 4]],
            [[0, 0, 4], [0, 0, 4]],
        ])
        c = b.colors('x', colormap='inferno')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

        # Viridis colormap
        t0 = np.array([  # Deepest array is a pixel
            [[68, 1, 84], [65, 68, 136]],
            [[42, 121, 143], [35, 170, 132]],
            [[125, 211, 80], [254, 232, 37]],
        ])
        t1 = np.array([
            [[254, 232, 37], [125, 211, 80]],
            [[35, 170, 132], [42, 121, 143]],
            [[65, 68, 136], [68, 1, 84]],
        ])
        t2 = np.array([
            [[68, 1, 84], [68, 1, 84]],
            [[68, 1, 84], [68, 1, 84]],
            [[68, 1, 84], [68, 1, 84]],
        ])
        c = b.colors('x', colormap='viridis')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))
        self.assertTrue(np.all(c[2] == t2))

    def test_colors_multiplier(self):
        # Test using the multiplier argument in colors

        w, h = 2, 3
        time = [1, 2, 3]
        b = myokit.DataBlock2d(w, h, time)
        x = np.array([  # Each 3 by 2 array is a point in time
            [[0, 1],
             [2, 3],
             [4, 5]],

            [[5, 4],
             [3, 2],
             [1, 0]],

            [[0, 0],
             [0, 0],
             [0, 0]],
        ])
        b.set2d('x', x)

        # Test with multiplier of 1 or less
        p = [255, 255, 255]
        q = [255, 204, 204]
        r = [255, 153, 153]
        s = [255, 102, 102]
        t = [255, 50, 50]
        u = [255, 0, 0]
        t0 = np.array([  # Deepest array is a pixel
            [p, q],
            [r, s],
            [t, u],
        ])
        t1 = np.array([
            [u, t],
            [s, r],
            [q, p],
        ])
        t2 = np.array([
            [p, p],
            [p, p],
            [p, p],
        ])
        c = b.colors('x', colormap='red', multiplier=1)
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))
        c = b.colors('x', colormap='red', multiplier=1.5)
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))
        c = b.colors('x', colormap='red', multiplier=0.5)
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

        t0 = np.array([
            [p, p, q, q],
            [p, p, q, q],
            [r, r, s, s],
            [r, r, s, s],
            [t, t, u, u],
            [t, t, u, u],
        ])
        t1 = np.array([
            [u, u, t, t],
            [u, u, t, t],
            [s, s, r, r],
            [s, s, r, r],
            [q, q, p, p],
            [q, q, p, p],
        ])
        t2 = np.array([
            [p, p, p, p],
            [p, p, p, p],
            [p, p, p, p],
            [p, p, p, p],
            [p, p, p, p],
            [p, p, p, p],
        ])
        c = b.colors('x', colormap='red', multiplier=2)
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

        t0 = np.array([
            [p, p, p, q, q, q],
            [p, p, p, q, q, q],
            [p, p, p, q, q, q],
            [r, r, r, s, s, s],
            [r, r, r, s, s, s],
            [r, r, r, s, s, s],
            [t, t, t, u, u, u],
            [t, t, t, u, u, u],
            [t, t, t, u, u, u],
        ])
        t1 = np.array([
            [u, u, u, t, t, t],
            [u, u, u, t, t, t],
            [u, u, u, t, t, t],
            [s, s, s, r, r, r],
            [s, s, s, r, r, r],
            [s, s, s, r, r, r],
            [q, q, q, p, p, p],
            [q, q, q, p, p, p],
            [q, q, q, p, p, p],
        ])
        t2 = np.array([
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
            [p, p, p, p, p, p],
        ])
        c = b.colors('x', colormap='red', multiplier=3)
        self.assertTrue(np.all(c == np.array([t0, t1, t2])))

    def test_combined(self):
        # Test loading, saving, conversion from data log.

        # Create simulation log with 1d data
        log = myokit.DataLog()
        log.set_time_key('time')
        t = np.linspace(0, 1, 20)
        log['time'] = t
        log['0.0.x'] = np.sin(t)
        log['0.1.x'] = np.cos(t)
        log['0.2.x'] = np.tan(t)
        log['1.0.x'] = np.sin(2 * t)
        log['1.1.x'] = np.cos(2 * t)
        log['1.2.x'] = np.tan(2 * t)

        # Convert to datablock
        b = log.block2d()

        # Check block contents
        self.assertTrue(np.all(b.time() == t))
        self.assertFalse(b.time() is t)
        self.assertEqual(b.len0d(), 0)
        self.assertEqual(b.len2d(), 1)
        x = b.get2d('x')
        for xx in range(2):
            for yy in range(3):
                self.assertTrue(np.all(x[:, yy, xx] == log['x', xx, yy]))

        # Make bigger log, try again
        log['pace'] = np.ones(t.shape) + t**2
        log['0.0.y'] = np.sqrt(t)
        log['0.1.y'] = 1 + np.sqrt(t)
        log['0.2.y'] = 2 + np.sqrt(t)
        log['1.0.y'] = np.sqrt(t * 2)
        log['1.1.y'] = 1 + np.sqrt(t * 2)
        log['1.2.y'] = 2 + np.sqrt(t * 3)

        # Convert to datablock
        b = log.block2d()

        # Check block contents
        self.assertTrue(np.all(b.time() == t))
        self.assertFalse(b.time() is t)
        self.assertEqual(b.len0d(), 1)
        self.assertEqual(b.len2d(), 2)
        self.assertTrue(np.all(b.get0d('pace') == log['pace']))
        self.assertFalse(b.get0d('pace') is log['pace'])
        x = b.get2d('x')
        y = b.get2d('y')
        for xx in range(2):
            for yy in range(3):
                self.assertTrue(np.all(x[:, yy, xx] == log['x', xx, yy]))
                self.assertTrue(np.all(y[:, yy, xx] == log['y', xx, yy]))

        # Test reading and writing
        with TemporaryDirectory() as td:
            fname = td.path('block2d.zip')
            b.save(fname)
            c = myokit.DataBlock2d.load(fname)

            # Test block contents
            self.assertTrue(np.all(b.time() == c.time()))
            self.assertFalse(b.time() is c.time())
            self.assertEqual(c.len0d(), 1)
            self.assertEqual(c.len2d(), 2)
            self.assertTrue(np.all(b.get0d('pace') == c.get0d('pace')))
            self.assertFalse(b.get0d('pace') is c.get0d('pace'))
            xb = b.get2d('x')
            xc = c.get2d('x')
            self.assertTrue(np.all(xb == xc))
            self.assertFalse(xb is xc)
            yb = b.get2d('y')
            yc = c.get2d('y')
            self.assertTrue(np.all(yb == yc))
            self.assertFalse(yb is yc)

    def test_combine(self):
        # Test the combine() method.

        # Create first data block
        w1, h1 = 2, 3
        t1 = [1, 2, 3]
        b1 = myokit.DataBlock2d(w1, h1, t1)
        x1 = np.array([  # Each 3 by 2 array is a point in time
            [[0, 1],
             [2, 3],
             [4, 5]],
            [[5, 4],
             [3, 2],
             [1, 0]],
            [[0, 0],
             [0, 0],
             [0, 0]],
        ])
        b1.set2d('x', x1)

        # Create first data block
        w2, h2 = 1, 2
        t2 = [1, 2, 3]
        b2 = myokit.DataBlock2d(w2, h2, t2)
        x2 = np.array([  # Each 2 by 1 array is a point in time
            [[0],
             [1]],
            [[2],
             [3]],
            [[4],
             [5]],
        ])
        b2.set2d('x', x2)

        # Basic combine
        m12 = {'x': ('x', 'x', -1)}
        b = myokit.DataBlock2d.combine(b1, b2, m12)
        c0 = [[0, 1, 0],
              [2, 3, 1],
              [4, 5, -1]]
        c1 = [[5, 4, 2],
              [3, 2, 3],
              [1, 0, -1]]
        c2 = [[0, 0, 4],
              [0, 0, 5],
              [0, 0, -1]]
        x = b.get2d('x')
        self.assertTrue(np.all(x[0] == c0))
        self.assertTrue(np.all(x[1] == c1))
        self.assertTrue(np.all(x[2] == c2))

        # Combine with automatic padding value
        mx = {'x': ('x', 'x')}
        b = myokit.DataBlock2d.combine(b1, b2, mx)
        c0 = [[0, 1, 0],
              [2, 3, 1],
              [4, 5, 0]]
        c1 = [[5, 4, 2],
              [3, 2, 3],
              [1, 0, 0]]
        c2 = [[0, 0, 4],
              [0, 0, 5],
              [0, 0, 0]]
        x = b.get2d('x')
        self.assertTrue(np.all(x[0] == c0))
        self.assertTrue(np.all(x[1] == c1))
        self.assertTrue(np.all(x[2] == c2))

        # Combine with pos1 and pos2 set
        m12 = {'x': ('x', 'x', -1)}
        b = myokit.DataBlock2d.combine(b1, b2, m12, pos1=(1, 0), pos2=(0, 1))
        c0 = [[-1, 0, 1],
              [0, 2, 3],
              [1, 4, 5]]
        c1 = [[-1, 5, 4],
              [2, 3, 2],
              [3, 1, 0]]
        c2 = [[-1, 0, 0],
              [4, 0, 0],
              [5, 0, 0]]
        x = b.get2d('x')
        self.assertTrue(np.all(x[0] == c0))
        self.assertTrue(np.all(x[1] == c1))
        self.assertTrue(np.all(x[2] == c2))

        # Combine with 0d information
        b1.set0d('y', [0, 10, 20])
        b2.set0d('y', [20, 30, 40])
        b = myokit.DataBlock2d.combine(b1, b2, m12, {'y': ('y', None)})
        self.assertTrue(np.all(b.get0d('y') == [0, 10, 20]))
        b = myokit.DataBlock2d.combine(b1, b2, m12, {'y': (None, 'y')})
        self.assertTrue(np.all(b.get0d('y') == [20, 30, 40]))
        self.assertRaises(
            ValueError, myokit.DataBlock2d.combine, b1, b2, m12,
            {'y': ('y', 'y')})

        # Test bad time points
        b1 = myokit.DataBlock2d(w1, h1, t1)
        b2 = myokit.DataBlock2d(w2, h2, [4, 5, 6])
        self.assertRaises(ValueError, myokit.DataBlock2d.combine, b1, b2, m12)

        # Test negative pos1, bad pos2
        b1 = myokit.DataBlock2d(w1, h1, t1)
        b2 = myokit.DataBlock2d(w2, h2, t2)
        self.assertRaises(
            ValueError, myokit.DataBlock2d.combine, b1, b2, m12, pos1=(0, -1))
        self.assertRaises(
            ValueError, myokit.DataBlock2d.combine, b1, b2, m12, pos2=(0, -1))

        # Test overlapping pos1, bad pos2
        b1 = myokit.DataBlock2d(w1, h1, t1)
        b2 = myokit.DataBlock2d(w2, h2, t2)
        self.assertRaises(
            ValueError, myokit.DataBlock2d.combine, b1, b2, m12, pos1=(0, 0),
            pos2=(1, 1))

    def test_from_data_log(self):
        # Test some edge cases of `DataBlock2d.from_log`.

        # Test valid construction
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = [1, 2, 3]
        d['x', 0, 0] = [0, 0, 0]
        d['x', 1, 0] = [1, 1, 2]
        d['x', 0, 1] = [1, 2, 2]
        d['x', 1, 1] = [3, 4, 5]
        d['x', 0, 2] = [6, 6, 6]
        d['x', 1, 2] = [7, 7, 2]
        myokit.DataBlock2d.from_log(d)

        # Deprecated alias
        with WarningCollector() as wc:
            myokit.DataBlock2d.from_DataLog(d)
        self.assertIn('deprecated', wc.text())

        # No time key
        d.set_time_key(None)
        self.assertRaises(ValueError, myokit.DataBlock2d.from_log, d)

        # Bad time key
        d.set_time_key('z')

        # Multi-dimensional time key
        d.set_time_key('0.0.x')
        self.assertRaises(ValueError, myokit.DataBlock2d.from_log, d)

        # 1-dimensional stuff
        d.set_time_key('time')
        myokit.DataBlock2d.from_log(d)
        d['y', 0] = [10, 10, 10]
        d['y', 1] = [20, 20, 20]
        self.assertRaises(ValueError, myokit.DataBlock2d.from_log, d)

        # Mismatched dimensions
        d = myokit.DataLog()
        d.set_time_key('time')
        d['time'] = [1, 2, 3]
        d['x', 0, 0] = [0, 0, 0]
        d['x', 1, 0] = [1, 1, 2]
        d['x', 0, 1] = [1, 2, 2]
        d['x', 1, 1] = [3, 4, 5]
        d['x', 0, 2] = [6, 6, 6]
        d['x', 1, 2] = [7, 7, 2]
        d['y', 0, 0] = [0, 0, 0]
        d['y', 1, 0] = [1, 1, 2]
        d['y', 0, 1] = [1, 2, 2]
        d['y', 1, 1] = [3, 4, 5]
        d['y', 0, 2] = [6, 6, 6]
        d['y', 1, 2] = [7, 7, 2]
        myokit.DataBlock2d.from_log(d)
        del d['0.2.y']
        del d['1.2.y']
        self.assertRaises(ValueError, myokit.DataBlock2d.from_log, d)

    def test_images(self):
        # Test the images() method.

        w, h = 2, 2
        time = [1, 2]
        b = myokit.DataBlock2d(w, h, time)
        x = np.array([  # Each 2 by 2 array is a point in time
            [[0, 1],
             [2, 3]],
            [[3, 4],
             [4, 5]],
        ])
        b.set2d('x', x)

        # Red colormap
        t0 = np.array([  # Like colors, but strided together (ARGB32)
            255, 255, 255, 255, 204, 204, 255, 255,
            153, 153, 255, 255, 102, 102, 255, 255
        ])
        t1 = np.array([
            102, 102, 255, 255, 50, 50, 255, 255,
            50, 50, 255, 255, 0, 0, 255, 255,
        ])

        c = b.images('x', colormap='red')
        self.assertTrue(np.all(c[0] == t0))
        self.assertTrue(np.all(c[1] == t1))

    def test_is_square(self):
        # Test the is_square method.

        b = myokit.DataBlock2d(1, 1, [1, 2, 3])
        self.assertTrue(b.is_square())
        b = myokit.DataBlock2d(1, 2, [1, 2, 3])
        self.assertFalse(b.is_square())
        b = myokit.DataBlock2d(10, 10, [1, 2, 3])
        self.assertTrue(b.is_square())

    def test_keys_and_items(self):
        # Test the keys0d and keys2d methods, plus the items methods.

        w = 2
        h = 3
        time = [1, 2, 3]
        b = myokit.DataBlock2d(w, h, time)

        # Test keys0d
        self.assertEqual(len(list(b.keys0d())), 0)
        pace = np.array([0, 0, 2])
        b.set0d('pace', pace)
        self.assertEqual(len(list(b.keys0d())), 1)
        self.assertEqual(len(dict(b.items0d())), 1)
        self.assertIn('pace', b.keys0d())
        b.set0d('poos', pace)
        self.assertEqual(len(list(b.keys0d())), 2)
        self.assertEqual(len(dict(b.items0d())), 2)
        self.assertIn('poos', b.keys0d())
        self.assertIn('poos', dict(b.items0d()).keys())

        # Test keys2d
        self.assertEqual(len(list(b.keys2d())), 0)
        self.assertEqual(len(dict(b.items2d())), 0)
        x = np.array([
            [[1, 2], [3, 4], [5, 6]],
            [[3, 4], [5, 6], [7, 8]],
            [[1, 1], [2, 2], [3, 3]]
        ])
        b.set2d('x', x)
        self.assertEqual(len(list(b.keys2d())), 1)
        self.assertEqual(len(dict(b.items2d())), 1)
        y = 3 * x + 1
        b.set2d('y', y)
        self.assertEqual(len(list(b.keys2d())), 2)
        self.assertEqual(len(dict(b.items2d())), 2)

    def test_dominant_eigenvalues(self):
        # Test the dominant_eigenvalues method.

        # Won't work on non-square matrix
        b = myokit.DataBlock2d(1, 2, [1])
        self.assertRaises(Exception, b.dominant_eigenvalues, 'x')

        # Proper test
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[2, 0, 0], [0, 3, 4], [0, 4, 9]]
        b.set2d('x', [x])
        self.assertEqual(b.dominant_eigenvalues('x')[0], 11)

        # Complex values
        # Note: This example has complex eigenvalues, but they're not the
        # dominant ones. If they were, they'd have the same magnitude so the
        # dominant one would be undefined (and so implementation dependent,
        # which causes issues for this test. See #230).
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[0.8, -0.6, 0], [0.6, 0.8, 0], [1, 2, 2]]
        b.set2d('x', [x])
        self.assertAlmostEqual(
            b.dominant_eigenvalues('x')[0], 2 + 0j)

    def test_eigenvalues(self):
        # Test the eigenvalues method.

        # Won't work on non-square matrix
        b = myokit.DataBlock2d(1, 2, [1])
        self.assertRaises(Exception, b.eigenvalues, 'x')

        # Proper test
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[2, 0, 0], [0, 3, 4], [0, 4, 9]]
        b.set2d('x', [x])
        self.assertTrue(np.all(b.eigenvalues('x') == [[11, 1, 2]]))

        # Complex values
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[0, 1, 0], [0, 0, 1], [1, 0, 0]]
        b.set2d('x', [x])
        e = b.eigenvalues('x')[0]
        self.assertAlmostEqual(e[0], -0.5 - np.sqrt(3) / 2j)
        self.assertAlmostEqual(e[1], -0.5 + np.sqrt(3) / 2j)
        self.assertAlmostEqual(e[2], 1)

    def test_largest_eigenvalues(self):
        # Test the largest_eigenvalues method.

        # Won't work on non-square matrix
        b = myokit.DataBlock2d(1, 2, [1])
        self.assertRaises(Exception, b.largest_eigenvalues, 'x')

        # Proper test
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[2, 0, 0], [0, 3, 4], [0, 4, 9]]
        b.set2d('x', [x])
        self.assertEqual(b.largest_eigenvalues('x')[0], 11)

        # Complex values
        b = myokit.DataBlock2d(3, 3, [1])
        x = [[0, 1, 0], [0, 0, 1], [1, 0, 0]]
        b.set2d('x', [x])
        self.assertAlmostEqual(b.largest_eigenvalues('x')[0], 1)

    def test_load_bad_file(self):
        # Test loading errors.

        # Not enough files
        path = os.path.join(DIR_IO, 'bad2d-1-not-enough-files.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Not enough files',
            myokit.DataBlock2d.load, path)

        # No header file
        path = os.path.join(DIR_IO, 'bad2d-2-no-header.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Header not found',
            myokit.DataBlock2d.load, path)

        # No data file
        path = os.path.join(DIR_IO, 'bad2d-3-no-data.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Data not found',
            myokit.DataBlock2d.load, path)

        # Not a zip
        path = os.path.join(DIR_IO, 'bad2d-4-not-a-zip.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Bad zip',
            myokit.DataBlock2d.load, path)

        # Unknown data type in data
        path = os.path.join(DIR_IO, 'bad2d-5-bad-data-type.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'Unrecognized data type',
            myokit.DataBlock2d.load, path)

        # Unknown data type in data
        path = os.path.join(DIR_IO, 'bad2d-6-time-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock2d.load, path)

        # Unknown data type in data
        path = os.path.join(DIR_IO, 'bad2d-7-0d-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock2d.load, path)

        # Unknown data type in data
        path = os.path.join(DIR_IO, 'bad2d-8-2d-too-short.zip')
        self.assertRaisesRegex(
            myokit.DataBlockReadError, 'larger data',
            myokit.DataBlock2d.load, path)

        # Test progress reporter
        path = os.path.join(DIR_IO, 'block2d.zip')
        p = TestReporter()
        self.assertFalse(p.entered)
        self.assertFalse(p.exited)
        self.assertFalse(p.updated)
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNotNone(b)
        self.assertTrue(p.entered)
        self.assertTrue(p.exited)
        self.assertTrue(p.updated)

        # Test cancelling in progress reporter
        p = CancellingReporter()
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(1)
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(2)
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNone(b)
        p = CancellingReporter(3)
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNone(b)

        # Test loading 1d block with 2d method
        path = os.path.join(DIR_DATA, 'cv1d.zip')
        b = myokit.DataBlock2d.load(path)
        self.assertIsInstance(b, myokit.DataBlock2d)
        # Test if None is passed through
        p = CancellingReporter()
        b = myokit.DataBlock2d.load(path, p)
        self.assertIsNone(b)

    def test_save_frame_csv(self):
        # Test the save_frame_csv() method.

        w, h = 2, 3
        time = [1, 2, 3]
        b = myokit.DataBlock2d(w, h, time)
        x = np.array([  # Each 3 by 2 array is a point in time
            [[0, 1],
             [2, 3],
             [4, 5]],

            [[5, 4],
             [3, 2],
             [1, 0]],

            [[0, 0],
             [0, 0],
             [0, 0]],
        ])
        b.set2d('x', x)

        with TemporaryDirectory() as d:
            path = d.path('test.csv')
            b.save_frame_csv(path, 'x', 0)
            with open(path, 'r') as f:
                lines = [str(x) for x in f.readlines()]
            self.assertEqual(lines[0], '"x","y","value"\n')
            self.assertEqual(lines[1], '0,0,0\n')
            self.assertEqual(lines[2], '1,0,1\n')
            self.assertEqual(lines[3], '0,1,2\n')
            self.assertEqual(lines[4], '1,1,3\n')
            self.assertEqual(lines[5], '0,2,4\n')
            self.assertEqual(lines[6], '1,2,5')

    def test_save_frame_grid(self):
        # Test the save_frame_grid() method.

        w, h = 2, 3
        time = [1, 2, 3]
        b = myokit.DataBlock2d(w, h, time)
        x = np.array([  # Each 3 by 2 array is a point in time
            [[0, 1],
             [2, 3],
             [4, 5]],

            [[5, 4],
             [3, 2],
             [1, 0]],

            [[0, 0],
             [0, 0],
             [0, 0]],
        ])
        b.set2d('x', x)

        with TemporaryDirectory() as d:
            path = d.path('test.csv')
            b.save_frame_grid(path, 'x', 0)
            with open(path, 'r') as f:
                lines = [str(x) for x in f.readlines()]
            self.assertEqual(lines[0], '0 1\n')
            self.assertEqual(lines[1], '2 3\n')
            self.assertEqual(lines[2], '4 5')

    def test_remove0d(self):
        # Test remove0d().

        # Add new 0d field and remove again
        b = myokit.DataBlock2d(2, 3, [1, 2, 3])
        self.assertEqual(b.len0d(), 0)
        b.set0d('pace', np.array([0, 0, 2]))
        self.assertEqual(b.len0d(), 1)
        b.remove0d('pace')
        self.assertEqual(b.len0d(), 0)

        # Time can't be removed
        self.assertRaises(KeyError, b.remove0d, 'time')

    def test_remove2d(self):
        # Test remove2d().

        # Add new 0d field and remove again
        b = myokit.DataBlock2d(2, 3, [1, 2, 3])
        x = np.array([
            [[1, 2], [3, 4], [5, 6]],
            [[3, 4], [5, 6], [7, 8]],
            [[1, 1], [2, 2], [3, 3]]
        ])
        self.assertEqual(b.len2d(), 0)
        b.set2d('x', np.array(x))
        self.assertEqual(b.len2d(), 1)
        b.remove2d('x')
        self.assertEqual(b.len2d(), 0)

    def test_set0d(self):
        # Test set0d().

        w = 2
        h = 3
        time = [1, 2, 3]
        b1 = myokit.DataBlock2d(w, h, time)

        # Test valid call
        pace = np.array([0, 0, 2])
        b1.set0d('pace', pace)
        self.assertTrue(np.all(b1.get0d('pace') == pace))

        # Test copying
        b1.set0d('pace', pace, copy=False)
        self.assertTrue(np.all(b1.get0d('pace') == pace))
        self.assertIs(b1.get0d('pace'), pace)
        b1.set0d('pace', pace, copy=True)
        self.assertTrue(np.all(b1.get0d('pace') == pace))
        self.assertIsNot(b1.get0d('pace'), pace)

        # Test bad calls
        self.assertRaises(ValueError, b1.set0d, '', pace)
        self.assertRaises(ValueError, b1.set0d, 'pace', [1, 2, 3, 4])
        self.assertRaises(ValueError, b1.set0d, 'pace', np.array([1, 2, 3, 4]))

    def test_set2d(self):
        # Test set2d().

        w = 2
        h = 3
        time = [1, 2, 3]
        b1 = myokit.DataBlock2d(w, h, time)

        # Test valid call
        x = np.array([
            [[1, 2], [3, 4], [5, 6]],
            [[3, 4], [5, 6], [7, 8]],
            [[1, 1], [2, 2], [3, 3]]
        ])

        b1.set2d('x', x)
        self.assertTrue(np.all(b1.get2d('x') == x))

        # Test copying
        b1.set2d('x', x, copy=False)
        self.assertTrue(np.all(b1.get2d('x') == x))
        self.assertIs(b1.get2d('x'), x)
        b1.set2d('x', x, copy=True)
        self.assertTrue(np.all(b1.get2d('x') == x))
        self.assertIsNot(b1.get2d('x'), x)

        # Test bad calls
        self.assertRaises(ValueError, b1.set2d, '', x)
        x = np.array([
            [[1, 2], [3, 4]],
            [[3, 4], [5, 6]],
            [[1, 1], [2, 2]],
        ])
        self.assertRaises(ValueError, b1.set2d, 'x', x)

    def test_to_log(self):
        # Test the `DataBlock2d.to_log()` method.

        # Create test block
        b = myokit.DataBlock2d(3, 2, [1, 2, 3, 4])
        b.set0d('pace', [0, 1, 0, 0])
        b.set2d(
            'voltage',
            [[[2, 1, 0],
              [1, 0, 0]],
             [[1, 2, 1],
              [2, 1, 0]],
             [[0, 1, 2],
              [1, 2, 1]],
             [[0, 0, 1],
              [0, 1, 2]]]
        )

        # Convert and inspect
        d = b.to_log()
        self.assertIn('time', d)
        self.assertIn('pace', d)
        self.assertIn('0.0.voltage', d)
        self.assertIn('1.0.voltage', d)
        self.assertIn('2.0.voltage', d)
        self.assertIn('0.1.voltage', d)
        self.assertIn('1.1.voltage', d)
        self.assertIn('2.1.voltage', d)
        self.assertEqual(len(d), 8)
        self.assertEqual(list(d.time()), [1, 2, 3, 4])
        self.assertEqual(list(d['pace']), [0, 1, 0, 0])
        self.assertEqual(list(d['0.0.voltage']), [2, 1, 0, 0])
        self.assertEqual(list(d['1.0.voltage']), [1, 2, 1, 0])
        self.assertEqual(list(d['2.0.voltage']), [0, 1, 2, 1])
        self.assertEqual(list(d['0.1.voltage']), [1, 2, 1, 0])
        self.assertEqual(list(d['1.1.voltage']), [0, 1, 2, 1])
        self.assertEqual(list(d['2.1.voltage']), [0, 0, 1, 2])

    def test_trace(self):
        # Tests the trace() method

        w = 2
        h = 3
        time = [1, 2, 3]
        b1 = myokit.DataBlock2d(w, h, time)
        x = np.array([
            [[1, 2], [3, 4], [5, 6]],
            [[3, 4], [5, 6], [7, 8]],
            [[1, 1], [2, 2], [3, 3]]
        ])
        b1.set2d('x', x)
        self.assertTrue(np.all(b1.trace('x', 1, 1) == [4, 6, 2]))


class ColorMapTest(unittest.TestCase):
    """
    Tests extra methods from ColorMap class.
    """
    def test_exists(self):
        self.assertFalse(myokit.ColorMap.exists('michael'))
        self.assertTrue(myokit.ColorMap.exists('red'))

    def test_get(self):
        self.assertIsInstance(myokit.ColorMap.get('red'), myokit.ColorMap)
        self.assertRaises(KeyError, myokit.ColorMap.get, 'michael')

    def test_names(self):
        names = list(myokit.ColorMap.names())
        self.assertIn('red', names)

    def test_image(self):
        # Test the image() method, that returns a colormap representation

        # Red colormap
        m = np.array([  # Like colors, but strided together (ARGB32)
            0, 0, 255, 255,
            50, 50, 255, 255,
            102, 102, 255, 255,
            153, 153, 255, 255,
            204, 204, 255, 255,
            255, 255, 255, 255,
        ])

        c = myokit.ColorMap.image('red', 1, 6)
        self.assertTrue(np.all(c == m))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Tests shared infrastructure in myokit.formats.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats

from myokit.tests import TemporaryDirectory


class FormatsTest(unittest.TestCase):
    """
    Test infrastructure from the formats module.
    """

    def test_register_external_importer(self):
        # Tests registering an external importer.

        class TestImporter(myokit.formats.Importer):
            pass

        self.assertRaisesRegex(
            KeyError, 'Importer not found',
            myokit.formats.importer, 'testo')

        myokit.formats.register_external_importer('testo', TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)

        # Test overwriting
        myokit.formats.register_external_importer('testo', TestImporter)
        e = myokit.formats.importer('testo')
        self.assertIsInstance(e, TestImporter)

        # Test removing
        myokit.formats.register_external_importer('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Importer not found',
            myokit.formats.importer, 'testo')

    def test_register_external_exporter(self):
        # Tests registering an external exporter.

        class TestExporter(myokit.formats.Exporter):
            pass

        self.assertRaisesRegex(
            KeyError, 'Exporter not found',
            myokit.formats.exporter, 'testo')

        myokit.formats.register_external_exporter('testo', TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)

        # Test overwriting
        myokit.formats.register_external_exporter('testo', TestExporter)
        e = myokit.formats.exporter('testo')
        self.assertIsInstance(e, TestExporter)

        # Test removing
        myokit.formats.register_external_exporter('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Exporter not found',
            myokit.formats.exporter, 'testo')

    def test_register_external_ewriter(self):
        # Tests registering an external expression writer.

        class TestExpressionWriter(myokit.formats.ExpressionWriter):
            pass

        self.assertRaisesRegex(
            KeyError, 'Expression writer not found',
            myokit.formats.ewriter, 'testo')

        myokit.formats.register_external_ewriter('testo', TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)

        # Test overwriting
        myokit.formats.register_external_ewriter('testo', TestExpressionWriter)
        e = myokit.formats.ewriter('testo')
        self.assertIsInstance(e, TestExpressionWriter)

        # Test removing
        myokit.formats.register_external_ewriter('testo', None)
        self.assertRaisesRegex(
            KeyError, 'Expression writer not found',
            myokit.formats.ewriter, 'testo')


class ExporterTest(unittest.TestCase):
    """
    Tests some parts of :class:`Exporter` that are not tested when testing the
    various exporter implementations.
    """
    def test_writable_dir(self):
        # Test :meth:`_test_writable_dir` for existing paths.

        m, p, x = myokit.load('example')
        e = myokit.formats.exporter('ansic')
        with TemporaryDirectory() as d:

            # Create in new dir
            path = d.path('new-dir')
            e.runnable(path, m, p)

            # Create again, in same dir (should be ok)
            e.runnable(path, m, p)

            # Create at location of file: not allowed!
            path = d.path('file')
            with open(path, 'w') as f:
                f.write('Hello')
            self.assertRaisesRegex(
                myokit.ExportError, 'file exists', e.runnable, path, m, p)


if __name__ == '__main__':
    unittest.main()



#
# Exports to CellML.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import myokit.lib.guess

import warnings


class CellMLExporter(myokit.formats.Exporter):
    """
    This:class:`Exporter <myokit.formats.Exporter>` creates a CellML model.
    """
    def __init__(self):
        super().__init__()

    def model(self, path, model, protocol=None, version='1.0'):
        """
        Writes a CellML model to the given filename.

        Arguments:

        ``path``
            The path/filename to write the generated code to.
        ``model``
            The model to export
        ``protocol``
            If given, an attempt will be made to convert the protocol to an
            expression and insert it into the model before exporting. See
            :meth:`myokit.lib.guess.add_embedded_protocol()` for details.
        ``version``
            The CellML version to write (1.0, 1.1, or 2.0).

        """
        # Load API and writer
        if version in ('1.0', '1.1'):
            import myokit.formats.cellml.v1 as cellml
        elif version in ('2.0'):
            import myokit.formats.cellml.v2 as cellml
        else:   # pragma: no cover
            raise ValueError(
                'Only versions 1.0, 1.1, and 2.0 are supported.')

        # Embed protocol
        if protocol is not None:
            model = model.clone()
            if not myokit.lib.guess.add_embedded_protocol(model, protocol):
                warnings.warn('Unable to embed stimulus protocol.')

        # Export
        cellml_model = cellml.Model.from_myokit_model(model, version)
        cellml.write_file(path, cellml_model)

    def supports_model(self):
        """
        Returns ``True``.
        """
        return True


class CellML1Exporter(CellMLExporter):
    """
    This:class:`Exporter <myokit.formats.Exporter>` creates a CellML 1.0 model.
    """
    def model(self, path, model, protocol=None):
        super().model(path, model, protocol, '1.0')


class CellML2Exporter(CellMLExporter):
    """
    This:class:`Exporter <myokit.formats.Exporter>` creates a CellML 2.0 model.
    """
    def model(self, path, model, protocol=None):
        super().model(path, model, protocol, '2.0')




#
# This module reads files in WCP format
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import struct
import numpy as np

import myokit

# Encoding of text portions of wcp file
_ENC = 'ascii'


class WcpFile(myokit.formats.SweepSource):
    """
    Represents a read-only WinWCP file (``.wcp``), stored at ``path``.

    Only files in the newer file format version 9 can be read. This version of
    the format was introduced in 2010. New versions of WinWCP can read older
    files and will convert them to the new format automatically when opened.

    WinWCP is a tool for recording electrophysiological data written by John
    Dempster of Strathclyde University. For more information, see
    https://documentation.help/WinWCP-V5.3.8/IDH_Topic750.htm

    WinWCP files contain a number of records ``NR``, each containing data from
    ``NC`` channels. Every channel has the same length, ``NP`` samples.
    Sampling happens at a fixed sampling rate.
    """
    def __init__(self, path):
        # The path to the file and its basename
        path = str(path)
        self._path = os.path.abspath(path)
        self._filename = os.path.basename(path)

        # Header info, per file, per record, and per channel
        self._version_str = None
        self._header = {}
        self._header_raw = {}
        self._record_headers = []
        self._channel_headers = []

        # Records
        self._records = []
        self._nr = None     # Records in file
        self._nc = None     # Channels per record
        self._np = None     # Samples per channel
        self._dt = None     # Sampling interval (s)

        # Channels
        self._channel_names = []
        self._channel_name_map = {}

        # Time signal
        self._time = None

        # Units
        self._time_unit = None
        self._channel_units = []
        self._unit_cache = {}

        # Read the file
        with open(path, 'rb') as f:
            self._read(f)

    def _read(self, f):
        """ Reads the file header & data. """
        # Header size is between 1024 and 16380, depending on number of
        # channels in the file following:
        #   n = (int((n_channels - 1)/8) + 1) * 1024
        # Read first part of header, determine version and number of channels
        # in the file
        data = f.read(1024).decode(_ENC)
        h = [x.strip().split('=') for x in data.split('\n')]
        h = dict([(x[0].lower(), x[1]) for x in h if len(x) == 2])
        if int(h['ver']) != 9:
            raise NotImplementedError(
                'Only able to read format version 9. Given file is in format'
                ' version ' + h['ver'])
        self._version_str = h['ver']

        # Get header size
        try:
            # Get number of 512 byte sectors in header
            #header_size = 512 * int(h['nbh'])
            # Seems to be size in bytes!
            header_size = int(h['nbh'])
        except KeyError:    # pragma: no cover
            # Calculate header size based on number of channels
            header_size = (int((int(h['nc']) - 1) / 8) + 1) * 1024

        # Read remaining header data
        if header_size > 1024:  # pragma: no cover
            data += f.read(header_size - 1024).decode(_ENC)
            h = [x.strip().split('=') for x in data.split('\n')]
            h = dict([(x[0].lower(), x[1]) for x in h if len(x) == 2])

        # Tidy up read data
        for k, v in h.items():
            # Convert to appropriate data type
            try:
                t = HEADER_FIELDS[k]
                if t == float:
                    # Allow for windows locale stuff
                    v = v.replace(',', '.')
                self._header[k] = t(v)
            except KeyError:
                self._header_raw[k] = v

        # Convert time
        # No, don't. It's in different formats depending on... user locale?
        # if 'ctime' in header:
        #    print(header['ctime'])
        #    ctime = time.strptime(header['ctime'], "%d/%m/%Y %H:%M:%S")
        #    header['ctime'] = time.strftime('%Y-%m-%d %H:%M:%S', ctime)

        # Get vital fields from header
        self._dt = self._header['dt']       # Sampling interval
        self._nr = self._header['nr']       # Records in file
        self._nc = self._header['nc']       # Channels per record
        try:
            self._np = self._header['np']   # Samples per channel
        except KeyError:
            self._np = (self._header['nbd'] * 512) // (2 * self._nc)

        # Get time units
        self._time_unit = myokit.units.s
        # Time as set by dt (which is what we need) is _always_ in seconds.
        # John Dempster says: "The TU= value referred to the time units which
        # were displayed in early versions of WinWCP and no longer exists in
        # recent WCP data files." (Email to michael, 2023-07-12).

        # Get channel-specific fields
        for i in range(self._nc):
            c = {}
            for k, t in HEADER_CHANNEL_FIELDS.items():
                c[k] = t(h[k + str(i)])
            self._channel_headers.append(c)
            self._channel_names.append(c['yn'])
            self._channel_name_map[c['yn']] = i
            self._channel_units.append(self._unit(c['yu']))

        # Analysis block size and data block size
        # Data is stored as 16 bit integers (little-endian)
        try:
            rab_size = 512 * self._header['nba']
        except KeyError:    # pragma: no cover
            rab_size = header_size
        try:
            rdb_size = 512 * self._header['nbd']
        except KeyError:    # pragma: no cover
            rdb_size = 2 * self._nc * self._np

        # Maximum A/D sample value at vmax
        adcmax = self._header['adcmax']

        # Read data records
        offset = header_size
        for i in range(self._nr):
            # Read analysis block
            f.seek(offset)

            # Status of signal (Accepted or rejected, as string)
            rstatus = f.read(8).decode(_ENC)

            # Type of recording, as string
            rtype = f.read(4).decode(_ENC)

            # Leak subtraction group number (float set by the user)
            group_number = struct.unpack(str('<f'), f.read(4))[0]

            # Time of recording, as float, not sure how to interpret
            rtime = struct.unpack(str('<f'), f.read(4))[0]

            # Sampling interval
            # It is technically possible to have different dts for different
            # records (see email J.D. 2023-07-12), but rare.
            # Not supported here!
            rint = round(struct.unpack(str('<f'), f.read(4))[0], 6)
            if rint != self._dt:  # pragma: no cover
                raise ValueError(
                    'Unsupported feature: WCP file contains more than one'
                    ' sampling rate.')

            # Maximum positive limit of A/D converter voltage range
            vmax = struct.unpack(
                str('<' + 'f' * self._nc), f.read(4 * self._nc))

            # String marker set by user
            marker = f.read(16).decode(_ENC).strip('\x00')

            # Store
            self._record_headers.append({
                'status': rstatus,
                'type': rtype,
                'rtime': rtime,
                'marker': marker,
            })

            # Delete unused
            del rstatus, rtype, group_number, rtime, rint, marker

            # Increase offset beyond analysis block
            offset += rab_size

            # Get data from data block
            data = np.memmap(
                self._path, np.dtype('<i2'), 'r',
                shape=(self._np, self._nc),
                offset=offset,
            )

            # Separate channels and apply scaling
            record = [
                vmx / (adcmax * h['yg']) * data[:, h['yo']].astype('f4')
                for h, vmx in zip(self._channel_headers, vmax)]

            self._records.append(record)

            # Increase offset beyong data block
            offset += rdb_size

        # Create time signal
        self._time = np.arange(self._np) * self._dt

    def __getitem__(self, key):
        return self._records[key]

    def __iter__(self):
        return iter(self._records)

    def __len__(self):
        return self._nr

    def _channel_id(self, channel_id):
        """ Checks an int or str channel id and returns a valid int. """
        if self._nr == 0:  # pragma: no cover
            raise KeyError(f'Channel {channel_id} not found (empty file).')

        # Handle string
        if isinstance(channel_id, str):
            return self._channel_name_map[channel_id]  # Pass KeyError to user

        int_id = int(channel_id)  # TypeError for user
        if int_id < 0 or int_id >= self._nc:
            raise IndexError(f'channel_id out of range: {channel_id}')
        return int_id

    def channel(self, channel_id, join_sweeps=False):
        # Docstring in SweepSource
        channel_id = self._channel_id(channel_id)
        time, data = [], []
        for r, h in zip(self._records, self._record_headers):
            time.append(self._time + h['rtime'])
            data.append(r[channel_id])
        if join_sweeps:
            return (np.concatenate(time), np.concatenate(data))
        return time, data

    def channels(self):
        """ Deprecated alias of :meth:`channel_count`. """
        # Deprecated since 2023-06-22
        import warnings
        warnings.warn(
            'The method `channels` is deprecated. Please use'
            ' WcpFile.channel_count() instead.')
        return self._nc

    def channel_count(self):
        # Docstring in SweepSource
        return self._nc

    def channel_names(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._channel_names)
        return self._channel_names[index]

    def channel_units(self, index=None):
        # Docstring in SweepSource
        if index is None:
            return list(self._channel_units)
        return self._channel_units[index]

    def da_count(self):
        # Docstring in SweepSource. Rest is allowed raise NotImplementedError
        return 0

    def equal_length_sweeps(self):
        # Docstring in SweepSource
        return True

    def filename(self):
        """ Returns this file's name. """
        return self._filename

    def info(self):
        """ Deprecated alias of :meth:`meta_str` """
        # Deprecated since 2023-07-12
        import warnings
        warnings.warn(
            'The method `info` is deprecated. Please use `meta_str` instead.')
        return self.meta_str(False)

    def log(self, join_sweeps=False, use_names=False, include_da=True):
        # Docstring in SweepSource

        # Create log
        log = myokit.DataLog()
        if self._nr == 0:  # pragma: no cover
            return log

        # Get channel names
        channel_names = self._channel_names
        if not use_names:
            channel_names = [f'{i}.channel' for i in range(self._nc)]

        # Gather data and return
        if join_sweeps:
            log['time'] = np.concatenate(
                [self._time + h['rtime'] for h in self._record_headers])
            for c, name in enumerate(channel_names):
                log[name] = np.concatenate([r[c] for r in self._records])
        else:
            # Return individual sweeps
            log['time'] = self._time
            for r, record in enumerate(self._records):
                for c, name in enumerate(channel_names):
                    log[name, r] = record[c]

        log.set_time_key('time')
        return log

    def matplotlib_figure(self):
        """ Creates and returns a matplotlib figure with this file's data. """
        import matplotlib.pyplot as plt
        f = plt.figure()
        axes = [f.add_subplot(self._nc, 1, 1 + i) for i in range(self._nc)]
        for record in self._records:
            for ax, channel in zip(axes, record):
                ax.plot(self._time, channel)
        return f

    def meta_str(self, verbose=False):
        # Docstring in SweepSource
        h = self._header
        out = []

        # Add file info
        out.append(f'WinWCP file: {self._filename}')
        out.append(f'WinWCP Format version {self._version_str}')
        out.append(f'Recorded on: {h["rtime"]}')

        # Basic records info
        out.append(f'  Number of records: {self._nr}')
        out.append(f'  Channels per record: {self._nc}')
        out.append(f'  Samples per channel: {self._np}')
        out.append(f'  Sampling interval: {self._dt} s')

        # Channel info
        for c in self._channel_headers:
            out.append(f'A/D channel: {c["yn"]}')
            out.append(f'  Unit: {c["yu"]}')

        # Record info
        out.append(
            'Records: Type, Status, Sampling Interval, Start, Marker')
        for i, r in enumerate(self._record_headers):
            out.append(f'Record {i}: {r["type"]}, {r["status"]},'
                       f' {r["rtime"]}, "{r["marker"]}"')

        # Parsed and unparsed header
        if verbose:
            out.append(f'{"-" * 35} header {"-" * 35}')
            for k, v in self._header.items():
                out.append(f'{k}: {v}')

            out.append(f'{"-" * 33} raw header {"-" * 33}')
            for k, v in self._header_raw.items():
                out.append(f'{k}: {v}')

        return '\n'.join(out)

    def myokit_log(self):
        """
        Deprecated method. Please use ``WcpFile.log(use_names=True)`` instead.
        """
        # Deprecated since 2023-06-22
        import warnings
        warnings.warn(
            'The method `myokit_log` is deprecated. Please use'
            ' WcpFile.log(use_names=True) instead.')
        return self.log(self)

    def path(self):
        """ Returns the path to this file. """
        return self._path

    def plot(self):
        """
        Deprecated method, please use :meth:`matplotlib_figure()` instead.

        Creates and shows a matplotlib figure of all data in this file.
        """
        # Deprecated since 2023-06-22
        import warnings
        warnings.warn(
            'The method `plot` is deprecated. Please use'
            ' WcpFile.matplotlib_figure() instead.')

        import matplotlib.pyplot as plt
        self.matplotlib_figure()
        plt.show()

    def record_count(self):
        """ Alias of :meth:`sweep_count`. """
        return self._nr

    def records(self):
        """ Deprecated alias of :meth:`sweep_count`. """
        # Deprecated since 2023-06-22
        import warnings
        warnings.warn(
            'The method `records` is deprecated. Please use'
            ' WcpFile.record_count() instead.')

        return self._nr

    def sample_count(self):
        """ Returns the number of samples in each channel. """
        return self._np

    def sweep_count(self):
        # Docstring in SweepSource
        return self._nr

    def time_unit(self):
        # Docstring in SweepSource
        return self._time_unit

    def times(self):
        """ Returns the time points sampled at. """
        return np.array(self._time)

    def _unit(self, unit_string):
        """ Parses a unit string and returns a :class:`myokit.Unit`. """
        try:
            return self._unit_cache[unit_string]
        except KeyError:
            unit = myokit.parse_unit(unit_string)
            self._unit_cache[unit_string] = unit
            return unit

    def values(self, record, channel):
        """
        Returns the values of channel ``channel``, recorded in record
        ``record``.
        """
        return self._records[record][channel]

    def version(self):
        """ Returns this file's version, as a string. """
        return self._version_str


HEADER_FIELDS = {
    'ver': int,        # WCP data file format version number
    'ctime': str,      # Create date/time
    'rtime': str,      # Start of recording time
    'nc': int,         # No. of channels per record
    'nr': int,         # No. of records in the file.
    'nbh': int,        # No. of 512 byte sectors in file header block
    'nba': int,        # No. of 512 byte sectors in a record analysis block
    'nbd': int,        # No. of 512 byte sectors in a record data block
    'ad': float,       # A/D converter input voltage range (V)
    'adcmax': int,     # Maximum A/D sample value
    'np': int,         # No. of A/D samples per channel
    'dt': float,       # A/D sampling interval (s)
    'nz': int,         # No. of samples averaged to calculate a zero level.
    'id': str,         # Experiment identification line
}


HEADER_CHANNEL_FIELDS = {
    'yn': str,         # Channel name
    'yu': str,         # Channel units
    'yg': float,       # Channel gain factor mV/units
    'yz': int,         # Channel zero level (A/D bits)
    'yo': int,         # Channel offset into sample group in data block
    'yr': int,         # ADCZeroAt, probably for old files
}
#TODO: Find out if we need to do something with yz and yg
